#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import logging
import re
from typing import Any, Collection, Dict, List, Optional, Set, Union

from sqlalchemy import log, types as sqltypes, util
from sqlalchemy.dialects.mysql.base import _DecodingRow
from sqlalchemy.dialects.mysql.reflection import _re_compile
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.schema import Table

from starrocks.common.consts import TableConfigKey
from starrocks.common.params import (
    ColumnAggInfoKeyWithPrefix,
    SRKwargsPrefix,
    TableInfoKey,
    TableInfoKeyWithPrefix,
    TableKind,
    TableObjectInfoKey,
)
from starrocks.common.types import PartitionType, TableEngine
from starrocks.common.utils import SQLParseError, TableAttributeNormalizer

from .common import utils
from .drivers.parsers import parse_data_type, parse_mv_refresh_clause
from .engine.interfaces import (
    MySQLKeyType,
    ReflectedDistributionInfo,
    ReflectedMVState,
    ReflectedPartitionInfo,
    ReflectedRefreshInfo,
    ReflectedState,
    ReflectedTableKeyInfo,
    ReflectedViewState,
)


logger = logging.getLogger(__name__)


class StarRocksInspector(Inspector):
    """
    The StarRocksInspector provides a custom inspector for the StarRocks dialect,
    allowing for reflection of StarRocks-specific database objects like views.
    """
    def __init__(self, bind):
        super().__init__(bind)

    def get_view(self, view_name: str, schema: Optional[str] = None, **kwargs: Any) -> ReflectedViewState:
        """
        Retrieves information about a specific view.

        :param view_name: The name of the view to inspect.
        :param schema: The schema of the view; defaults to the default schema name if None.
        :param kwargs: Additional arguments passed to the dialect's get_view method.
        :return: A ReflectedViewState object.
        :raises: NoSuchTableError if the view does not exist.
        """
        return self.dialect.get_view(self.bind, view_name, schema=schema, **kwargs)

    def get_materialized_view(self, view_name: str, schema: Optional[str] = None, **kwargs: Any) -> Optional[ReflectedMVState]:
        """
        Retrieves information about a specific materialized view.

        :param view_name: The name of the materialized view to inspect.
        :param schema: The schema of the materialized view; defaults to the default schema name if None.
        :param kwargs: Additional arguments passed to the dialect's get_materialized_view method.
        :return: A ReflectedMVState object, or None if the materialized view does not exist.
        """
        return self.dialect.get_materialized_view(self.bind, view_name, schema=schema, **kwargs)

    def get_materialized_view_definition(self, view_name: str, schema: Optional[str] = None, **kwargs: Any) -> Optional[str]:
        """
        Retrieves the definition of a specific materialized view.

        :param view_name: The name of the materialized view to inspect.
        :param schema: The schema of the materialized view; defaults to the default schema name if None.
        :param kwargs: Additional arguments passed to the dialect's get_materialized_view_definition method.
        :return: The materialized view definition as a string, or None if the view does not exist.
        """
        mv_state = self.get_materialized_view(view_name, schema=schema, **kwargs)
        return mv_state.definition if mv_state else None

    def reflect_table(
        self,
        table: Table,
        include_columns: Optional[Collection[str]] = None,
        exclude_columns: Collection[str] = (),
        resolve_fks: bool = True,
        _extend_on: Optional[Set[Table]] = None,
        _reflect_info: Optional[Any] = None
    ) -> None:
        """
        Override to set VIEW/MV specific attributes.
        """

        # 1. Call parent class (will call get_pk_constraints, etc., which will trigger _setup_parser)
        #    And, it will set all dialect options from parsed_state, including for a View or a Mv or a Table.
        super().reflect_table(table, include_columns, exclude_columns, resolve_fks, _extend_on, _reflect_info)
        # comment is already set to Table.comment, the starrocks_comment is not used again.
        self._delete_comment_from_dialect_options(table)

        # 1. Get table_kind and parsed_state, which will use the cached parsed_state)
        parsed_state = self.dialect._parsed_state_or_create(self.bind, table.name, table.schema, info_cache=self.info_cache)
        table_kind = parsed_state.table_kind
        logger.debug("reflect %s: %s, parsed_state: %s.", table_kind, table.name, parsed_state)

        # 3. Set info['table_kind']
        table.info[TableObjectInfoKey.TABLE_KIND] = table_kind

        # 4. Set specific attributes based on type
        if table_kind == TableKind.VIEW:
            self._reflect_view_attributes(table, parsed_state)
        elif table_kind == TableKind.MATERIALIZED_VIEW:
            self._reflect_mv_attributes(table, parsed_state)

    @staticmethod
    def _delete_comment_from_dialect_options(table: Table):
        try:
            del table.dialect_kwargs[TableInfoKeyWithPrefix.COMMENT]
        except KeyError:
            # starrocks_comment does not exist in dialect_kwargs
            pass

    def _reflect_view_attributes(self, table: Table, view_state: ReflectedViewState) -> None:
        """Set View specific attributes from ReflectedViewState.

        Note: Column information (including comments) is already handled by
        SQLAlchemy's standard reflection flow via get_columns(), which returns
        ReflectedColumn dictionaries from view_state.columns.
        """
        table.info[TableObjectInfoKey.DEFINITION] = view_state.definition
        # if view_state.security:
        #     table.dialect_options.setdefault(DialectName, {})[TableInfoKey.SECURITY] = view_state.security

    def _reflect_mv_attributes(self, table, mv_state: ReflectedMVState):
        """Set MV specific attributes from ReflectedMVState"""
        table.info[TableObjectInfoKey.DEFINITION] = mv_state.definition
        # table.dialect_options.setdefault(DialectName, {}).update(mv_state.table_options)


@log.class_logger
class StarRocksTableDefinitionParser(object):
    """
    This parser is responsible for interpreting the raw data returned from
    StarRocks' `information_schema` and `SHOW` commands.

    For columns, the base attributes (name, type, nullable, default) are
    parsed here, leveraging the underlying MySQL dialect where possible.
    This dialect-specific implementation adds logic to parse StarRocks-specific
    attributes that are not present in standard MySQL, such as the aggregation
    type on a column (e.g., 'SUM', 'REPLACE', 'KEY'). This is achieved by
    querying `SHOW FULL COLUMNS` and processing the 'Extra' field.

    Other standard column attributes are assumed to be handled correctly by
    the base MySQL dialect's reflection mechanisms.

    MySQLTableDefinitionParser uses regex to parse information, so it's not
    used here.
    """

    _COLUMN_TYPE_PATTERN = re.compile(r"^(?P<type>\w+)(?:\s*\((?P<args>.*?)\))?\s*(?:(?P<attr>unsigned))?$")
    _TABLE_KEY_PATTERN = re.compile(r'\s*(\w+\s+KEY)\s*\((.*)\)\s*', re.IGNORECASE)
    _BUCKETS_PATTERN = re.compile(r'\sBUCKETS\s+(\d+)', re.IGNORECASE)
    _BUCKETS_REPLACE_PATTERN = re.compile(r'\s+BUCKETS\s+\d+', re.IGNORECASE)
    _PARTITION_BY_PATTERN = re.compile(r"PARTITION BY\s*(.+?)(?=\s*(?:DISTRIBUTED BY|ORDER BY|REFRESH|PROPERTIES|AS|\Z))", re.IGNORECASE | re.DOTALL)

    _VIEW_SECURITY_PATTERN = re.compile(r'\s+SECURITY\s+(INVOKER|DEFINER|NONE)\b', re.IGNORECASE)

    # Patterns to parse CREATE MATERIALIZED VIEW statement
    _MV_REFRESH_PATTERN = re.compile(r"\s*REFRESH\s+(.+?)(?=\s*(?:PARTITION BY|DISTRIBUTED BY|ORDER BY|PROPERTIES|AS|\Z))", re.IGNORECASE | re.DOTALL)
    _MV_PROPERTIES_PATTERN = re.compile(r"\s*PROPERTIES\s*\((.+?)\)(?=\s*(?:PARTITION BY|DISTRIBUTED BY|ORDER BY|REFRESH|AS|\Z))", re.IGNORECASE | re.DOTALL)
    _MV_AS_DEFINITION_PATTERN = re.compile(r"\s*AS\s*((?:WITH|SELECT)\s*.+)", re.IGNORECASE | re.DOTALL)

    def __init__(self, dialect, preparer):
        self.dialect = dialect
        self.preparer = preparer
        self._re_csv_int = _re_compile(r"\d+")
        self._re_csv_str = _re_compile(r"\x27(?:\x27\x27|[^\x27])*\x27")

    def parse_table(
        self,
        table: _DecodingRow,
        table_config: Dict[str, Any],
        columns: List[_DecodingRow],
        column_2_agg_type: Dict[str, str] = None,
        column_autoinc: dict[str, bool] = None,
        charset: str = None,
    ) -> ReflectedState:
        """
        Parses the raw reflection data into a structured ReflectedState object.

        :param table: A row from `information_schema.tables`.
        :param table_config: A dictionary representing a row from `information_schema.tables_config`,
                             augmented with the 'PARTITION_CLAUSE'.
        :param columns: A list of rows from `information_schema.columns`.
        :param column_2_agg_type: A dictionary mapping column names to their aggregation types.
        :param charset: The character set of the table.
        :return: A ReflectedState object containing the parsed table information.
        """
        # logger.debug("parsing table: %s", table.TABLE_NAME)
        sorted_columns = self._sort_columns(columns)
        reflected_table_info = ReflectedState(
            table_name=table.TABLE_NAME,
            columns=[
                self._parse_column(
                    column=column,
                    col_2_autoinc=column_autoinc,
                    **{ColumnAggInfoKeyWithPrefix.AGG_TYPE: column_2_agg_type.get(column.COLUMN_NAME)} if column_2_agg_type else {},
                )
                for column in sorted_columns
            ],
            table_options=self._parse_table_options(
                table.TABLE_NAME, schema=table.TABLE_SCHEMA,
                table=table, table_config=table_config, sorted_columns=sorted_columns
            ),
            keys=[{
                "type": self._get_mysql_key_type(table_config=table_config),
                "columns": [(c, None, None) for c in self._get_key_columns(sorted_columns=sorted_columns)],
                "parser": None,
                "name": None,
            }],
        )
        logger.debug("reflected table info for table: %r, info: '%s'", table.TABLE_NAME, reflected_table_info)
        return reflected_table_info

    def _parse_column(self, column: _DecodingRow, col_2_autoinc: dict | None = None, **kwargs: Any) -> dict:
        """
        Parse column from information_schema.columns table.
        It returns dictionary with column informations expected by sqlalchemy.

        Args:
            column: A row from `information_schema.columns`.
            col_2_autoinc: A dictionary of column names and their auto increment info.
                The key is the column name, the value is True/False whether the column has auto increment.
                Example: `{"col1": True, "col2": False, ...}`
            kwargs: Additional keyword arguments, with prefix `starrocks_`, passed to the dialect.
                currently only support:
                    - starrocks_is_agg_key: Whether the column is a key column.
                    - starrocks_agg_type: The aggregate type of the column.

        Returns:
            A dictionary with column information expected by sqlalchemy.
            It's the same as the `ReflectedColumn` object.
        """
        col_data = {
            "name": column.COLUMN_NAME,
            "type": self._parse_column_type(column=column),
            "nullable": column.IS_NULLABLE != "NO",
            "default": self._ColumnDefaultValueRenderer().build(column),
            "autoincrement": col_2_autoinc.get(column.COLUMN_NAME, False) if col_2_autoinc else False,
            "comment": column.COLUMN_COMMENT or None,
            "dialect_options": {
                k: v for k, v in kwargs.items() if v is not None
            }
        }
        if column.GENERATION_EXPRESSION:
            col_data["computed"] = {
                "sqltext": column.GENERATION_EXPRESSION
            }

        # logger.debug("parsed column data for column: %s is %s", column.COLUMN_NAME, col_data)

        return col_data

    class _ColumnDefaultValueRenderer:
        """
        Apply StarRocks-friendly default value rendering:
        1. DATETIME CURRENT_TIMESTAMP defaults are emitted verbatim.
        2. Function-looking defaults become `(func_expr)`.
        3. Remaining literals are double quoted.
        """
        def build(self, column: _DecodingRow) -> Optional[str]:
            raw_default = getattr(column, "COLUMN_DEFAULT", None)
            if raw_default is None:
                return None

            default = str(raw_default).strip()
            if default is None or default.upper() == "NULL":
                return None

            if self._is_datetime_column(column) and default.upper().startswith("CURRENT_TIMESTAMP"):
                return default

            if self._looks_like_function(default):
                return self._wrap_function_default(default)

            return f"{default!r}"

        @staticmethod
        def _is_datetime_column(column: _DecodingRow) -> bool:
            column_type = getattr(column, "COLUMN_TYPE", "") or ""
            return column_type.upper().startswith("DATETIME")

        @staticmethod
        def _looks_like_function(default: str) -> bool:
            return bool(re.match(r"^[A-Za-z_][A-Za-z0-9_]*\s*\(.*\)$", default))

        @staticmethod
        def _wrap_function_default(default: str) -> str:
            trimmed = default.strip()
            if trimmed.startswith("(") and trimmed.endswith(")"):
                return trimmed
            return f"({trimmed})"

    def _parse_column_type(self, column: _DecodingRow) -> Any:
        """
        Parse column type from information_schema.columns table.
        It splits column type into type and arguments.
        After that it creates instance of column type.

        Some special cases:
            - LARGEINT: treat 'bigint(20) unsigned' as 'LARGEINT'
        """
        try:
            return parse_data_type(column.COLUMN_TYPE)
        except Exception as e:
            logger.warning(f"Could not parse type string '{column.COLUMN_TYPE}' for column '{column.COLUMN_NAME}'. Error: {e}")
            match = self._COLUMN_TYPE_PATTERN.match(column.COLUMN_TYPE)
            if match:
                type_ = match.group("type")
            else:
                type_ = column.COLUMN_TYPE

            util.warn(
                "Did not recognize type '%s' of column %r" % (type_, column.COLUMN_NAME)
            )
            return sqltypes.NullType

    def _get_mysql_key_type(self, table_config: Dict[str, Any]) -> str:
        """
        Get key type from information_schema.tables_config table.
        And return the MySQL's key type, as MySQLKeyType
        But, directly return the MySQLKeyType.PRIMARY, for check only
        """
        # table_model_to_key_type_map: Dict[str, MySQLKeyType] = {
        #     TableModel.DUP_KEYS: MySQLKeyType.UNIQUE,
        #     TableModel.DUP_KEYS2: MySQLKeyType.UNIQUE,
        #     TableModel.AGG_KEYS: MySQLKeyType.UNIQUE,
        #     TableModel.AGG_KEYS2: MySQLKeyType.UNIQUE,
        #     TableModel.PRI_KEYS: MySQLKeyType.PRIMARY,
        #     TableModel.PRI_KEYS2: MySQLKeyType.PRIMARY,
        #     TableModel.UNQ_KEYS: MySQLKeyType.UNIQUE,
        #     TableModel.UNQ_KEYS2: MySQLKeyType.UNIQUE,
        # }
        # return str(table_model_to_key_type_map.get(table_config.get(TableConfigKey.TABLE_MODEL), "").value)
        return str(MySQLKeyType.PRIMARY.value)

    def _sort_columns(self, columns: List[_DecodingRow]) -> List[_DecodingRow]:
        """
        Sort columns by ORDINAL_POSITION.
        """
        return sorted(columns, key=lambda col: col.ORDINAL_POSITION)

    def _get_key_columns(self, sorted_columns: List[_DecodingRow]) -> List[str]:
        """
        Get list of key columns (COLUMN_KEY) from information_schema.columns table.
        It returns list of column names that are part of key.

        Currently, we can't extract the key columns from information_schema.tables_config.
        """
        return [c.COLUMN_NAME for c in sorted_columns if c.COLUMN_KEY]

    @staticmethod
    def parse_key_clause(table_key: str) -> Optional[ReflectedTableKeyInfo]:
        """
        Parses a raw TABLE KEY clause string into a structured ReflectedTableKeyInfo object.
        It's not used now.
        """
        if not table_key or table_key.strip() == "":
            return None
        key_match = StarRocksTableDefinitionParser._TABLE_KEY_PATTERN.search(table_key)
        if key_match:
            type = key_match.group(1)
            columns = key_match.group(2)
            return ReflectedTableKeyInfo(type=type, columns=columns)
        else:
            logger.error(f"Invalid table key clause: '{table_key}'")
            return None

    @staticmethod
    def parse_partition_clause(partition_clause: str) -> Optional[ReflectedPartitionInfo]:
        """
        Parses a raw PARTITION BY clause string into a structured ReflectedPartitionInfo object.

        This method handles RANGE, LIST, and expression partitioning schemes. It
        extracts the partition type, or expression used for partitioning,
        and any pre-defined partition clauses
        (e.g., `(PARTITION p1 VALUES LESS THAN ('100'), PARTITION p2 VALUES LESS THAN ('200'))`).

        Args:
            partition_clause: The raw string of the PARTITION BY clause from a
                `SHOW CREATE TABLE` statement.

        Returns:
            A `ReflectedPartitionInfo` object containing the parsed details.
        """
        if not partition_clause:
            return None

        clause_upper = partition_clause.strip().upper()
        partition_method: str
        pre_created_partitions: Optional[str] = None

        # Check for RANGE or LIST partitioning
        if clause_upper.startswith(PartitionType.RANGE) or clause_upper.startswith(PartitionType.LIST):
            partition_type = PartitionType.RANGE if clause_upper.startswith(PartitionType.RANGE) else PartitionType.LIST

            # Find the end of the RANGE/LIST(...) part using robust parenthesis matching
            open_paren_index = partition_clause.find('(')
            if open_paren_index != -1:
                close_paren_index = utils.find_matching_parenthesis(partition_clause, open_paren_index)
                if close_paren_index != -1:
                    partition_method = partition_clause[:close_paren_index + 1].strip()
                    rest = partition_clause[close_paren_index + 1:].strip()
                    if rest:
                        pre_created_partitions = rest
                else:  # Fallback for mismatched parentheses
                    raise SQLParseError(f"Invalid partition clause, mismatched parentheses: {partition_clause}")
            else:  # Fallback for no parentheses
                raise SQLParseError(f"Invalid partition clause, no columns specified: {partition_clause}")
        else:
            # If not RANGE or LIST, it's an expression-based partition
            # normalize partition method to be without outer parentheses
            partition_type = PartitionType.EXPRESSION
            partition_method = TableAttributeNormalizer.remove_outer_parentheses(partition_clause)

        return ReflectedPartitionInfo(
            type=partition_type,
            partition_method=partition_method,
            pre_created_partitions=pre_created_partitions
        )

    @staticmethod
    def parse_distribution_clause(distribution: str) -> Union[ReflectedDistributionInfo, None]:
        """Parse DISTRIBUTED BY string to extract distribution method and buckets.
        Args:
            distribution: String like "HASH(id) BUCKETS 8" or "HASH(id)"
        Returns:
            ReflectedDistributionInfo object
        """
        if not distribution:
            return None
        buckets_match = StarRocksTableDefinitionParser._BUCKETS_PATTERN.search(distribution)

        if buckets_match:
            buckets = int(buckets_match.group(1))
            # Remove BUCKETS part to get pure distribution
            distribution_method = StarRocksTableDefinitionParser._BUCKETS_REPLACE_PATTERN.sub('', distribution).strip()
        else:
            buckets = None
            distribution_method = distribution

        return ReflectedDistributionInfo(
            type=None,
            columns=None,
            distribution_method=distribution_method,
            buckets=buckets,
        )

    @staticmethod
    def parse_distribution(distribution: Optional[Union[ReflectedDistributionInfo, str]]
                           ) -> Union[ReflectedDistributionInfo, None]:
        if not distribution or isinstance(distribution, ReflectedDistributionInfo):
            return distribution
        return StarRocksTableDefinitionParser.parse_distribution_clause(distribution)

    def _get_distribution_info(self, table_config: dict[str, Any]) -> ReflectedDistributionInfo:
        """
        Get distribution from information_schema.tables_config table.
        It returns ReflectedDistributionInfo representation of distribution option.
        """
        return ReflectedDistributionInfo(
            type=table_config.get(TableConfigKey.DISTRIBUTE_TYPE),
            columns=table_config.get(TableConfigKey.DISTRIBUTE_KEY),
            distribution_method=None,
            buckets=table_config.get(TableConfigKey.DISTRIBUTE_BUCKET),
        )

    def _parse_common_table_options(self, table_kind: str, table_row: Optional[_DecodingRow]) -> Dict[str, Any]:
        """
        Parses common options from an information_schema.tables row.
        Args:
            table_kind: Table kind, such as 'TABLE' or 'MATERIALIZED VIEW'.
            table_row: A row from `information_schema.tables`.
        """
        opts = {}
        if not table_row:
            return opts
        if table_row.TABLE_COMMENT and table_row.TABLE_COMMENT != 'OLAP':
            # logger.debug("table.TABLE_COMMENT: %s", table_row.TABLE_COMMENT)
            if table_kind == TableKind.MATERIALIZED_VIEW and table_row.TABLE_COMMENT == 'MATERIALIZED_VIEW':
                pass  # SR will automatically set the COMMENT `MATERIALIZED_VIEW` for MV if not provided
            elif table_kind == TableKind.VIEW and table_row.TABLE_COMMENT == 'VIEW':
                pass  # SR will automatically set the COMMENT `VIEW` for VIEW if not provided
            else:
                opts[TableInfoKeyWithPrefix.COMMENT] = table_row.TABLE_COMMENT
        return opts

    def _parse_general_table_options(self, table_name: str, schema: Optional[str] = None, table_config: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Parses general table options from a `tables_config` dict (not the original _DecodingRow object).
        This logic is shared between table and materialized view reflection.
        """
        table_fqn = utils.gen_simple_qualified_name(table_name, schema)
        logger.debug("parse general table options for table: %r.", table_fqn)

        opts = {}
        if partition_clause := table_config.get(TableConfigKey.PARTITION_CLAUSE):
            # logger.debug("table_config.%s: %s", TableConfigKey.PARTITION_CLAUSE, partition_clause)
            opts[TableInfoKeyWithPrefix.PARTITION_BY] = self.parse_partition_clause(partition_clause)

        distribute_key = table_config.get(TableConfigKey.DISTRIBUTE_KEY)
        distribute_type = table_config.get(TableConfigKey.DISTRIBUTE_TYPE)
        if distribute_key or distribute_type:
            # logger.debug("table_config.%s: %s", TableConfigKey.DISTRIBUTE_KEY, distribute_key)
            # logger.debug("table_config.%s: %s", TableConfigKey.DISTRIBUTE_TYPE, distribute_type)
            opts[TableInfoKeyWithPrefix.DISTRIBUTED_BY] = str(self._get_distribution_info(table_config))

        if sort_key := table_config.get(TableConfigKey.SORT_KEY):
            # logger.debug("table_config.%s: %s", TableConfigKey.SORT_KEY, sort_key)
            opts[TableInfoKeyWithPrefix.ORDER_BY] = sort_key

        if properties := table_config.get(TableConfigKey.PROPERTIES):
            # logger.debug("table_config.%s: %s", TableConfigKey.PROPERTIES, properties)
            try:
                opts[TableInfoKeyWithPrefix.PROPERTIES] = dict(json.loads(properties or "{}").items())
            except json.JSONDecodeError:
                logger.info(f"properties are not valid JSON: {properties}")
        return opts

    def _parse_table_options(self, table_name: str, schema: Optional[str],
            table: _DecodingRow, table_config: Dict[str, Any], sorted_columns: List[_DecodingRow]
        ) -> Dict:
        """
        Parse table options from `information_schema` views,
        and generate the table options with `starrocks_` prefix, which will be used to reflect a Table().
        Then, these options will be exactly the same as the options of a sqlalchemy.Table()
        which is created by users manually, for both sqlalchemy.Table() or ORM styles.

        NOTE: partition info is extracted from SHOW CREATE TABLE rather than directly from information_schema.tables_config.
            But, here we can directly use table_config, bucause they are already filled into the table_config dict.

        Args:
            table: A row from `information_schema.tables`.
            table_config: A dictionary representing a row from `information_schema.tables_config`,
                             augmented with the 'PARTITION_CLAUSE'.
            sorted_columns: A list of rows from `information_schema.columns`, sorted by ORDINAL_POSITION.

        Returns:
            A dictionary of StarRocks-specific table options with the 'starrocks_' prefix.
        """
        opts = self._parse_common_table_options(TableKind.TABLE, table)
        table_fqn = utils.gen_simple_qualified_name(table_name, schema)
        logger.debug("parse table options for table: %r.", table_fqn)

        # Set partition, distribution, sort key, properties
        opts.update(self._parse_general_table_options(table_name, schema, table_config))

        if table_engine := table_config.get(TableConfigKey.TABLE_ENGINE):
            # logger.debug("table_config.%s: %s", TableConfigKey.TABLE_ENGINE, table_engine)
            # change CLOUD_NATIVE to OLAP for shared-data mode.
            if table_engine.upper() == TableEngine.CLOUD_NATIVE:
                table_engine = TableEngine.OLAP
            # if table_engine.upper() != TableEngine.OLAP:
            #     raise NotImplementedError(f"Table engine {table_engine} is not supported now.")
            opts[TableInfoKeyWithPrefix.ENGINE] = table_engine.upper()

        # Get key type from information_schema.tables_config.TABLE_MODEL,
        # and key columns from information_schema.columns.COLUMN_KEY
        if table_model := table_config.get(TableConfigKey.TABLE_MODEL):
            # logger.debug("table_config.%s: %s", TableConfigKey.TABLE_MODEL, table_model)
            # convert to key string, such as "PRIMARY_KEY", not PRIMARY KEY"
            key_str = TableInfoKey.MODEL_TO_KEY_MAP.get(table_model)
            if key_str:
                key_columns_str = ", ".join(self._get_key_columns(sorted_columns))
                prefixed_key = f"{SRKwargsPrefix}{key_str}"
                opts[prefixed_key] = key_columns_str

        return opts

    def parse_view(
        self,
        view_row: _DecodingRow,
        table_row: Optional[_DecodingRow],
        column_rows: List[_DecodingRow],
        create_view_sql: Optional[str] = None
    ) -> ReflectedViewState:
        """
        Parses raw reflection data into a structured ReflectedViewState object.

        Args:
            view_row: Row from information_schema.views
            table_row: Optional row from information_schema.tables (for comment)
            column_rows: Rows from information_schema.columns (for column names, types, and comments)
            create_view_sql: Optional CREATE VIEW statement from SHOW CREATE VIEW (to parse SECURITY)

        Returns:
            A ReflectedViewState object with parsed view information
        """
        state = ReflectedViewState(
            table_name=view_row.TABLE_NAME,
            definition=view_row.VIEW_DEFINITION,
        )

        # Parse columns using standard _parse_column method
        # For views, we care about name, type, and comment
        # Type and nullable are inferred from the SELECT statement
        if column_rows:
            sorted_columns = self._sort_columns(column_rows)
            state.columns = [self._parse_column(col) for col in sorted_columns]

        table_options = self._parse_common_table_options(TableKind.VIEW, table_row)

        # table_options[TableInfoKeyWithPrefix.SECURITY] = view_row.SECURITY_TYPE.upper()
        # Parse SECURITY from SHOW CREATE VIEW output
        # Note: information_schema.views.SECURITY_TYPE is always empty in StarRocks (v3.5)
        if security := self._parse_sql_security_from_create_view(create_view_sql):
            table_options[TableInfoKeyWithPrefix.SECURITY] = security

        state.table_options = table_options
        return state

    def _parse_sql_security_from_create_view(self, create_view_sql: Optional[str]) -> Optional[str]:
        """
        Parse SECURITY clause from CREATE VIEW statement.

        Args:
            create_view_sql: CREATE VIEW statement from SHOW CREATE VIEW

        Returns:
            'INVOKER' or 'DEFINER' if found, None otherwise

        Example:
            CREATE VIEW v1 SECURITY INVOKER AS SELECT ...
            -> Returns 'INVOKER'
        """
        if not create_view_sql:
            return None

        # Match: SECURITY {INVOKER|DEFINER|NONE}
        match = self._VIEW_SECURITY_PATTERN.search(create_view_sql)

        if match:
            security_type = match.group(1).upper()
            # logger.debug("Parsed SECURITY: %s from CREATE VIEW: %s", security_type, create_view_sql[:200])
            return security_type
        else:
            logger.debug("No SECURITY match in CREATE VIEW: %s", create_view_sql[:200] if create_view_sql else 'None')

        return None

    def parse_mv(
        self,
        mv_row: _DecodingRow,
        table_row: Optional[_DecodingRow],
        config_row: Optional[_DecodingRow],
    ) -> ReflectedMVState:
        """
        Parses all raw reflection data for a Materialized View into a ReflectedMVState.
        This is the main entry point for MV reflection parsing.

        Args:
            mv_row: A row from `information_schema.materialized_views`.
            table_row: An optional row from `information_schema.tables` for the same mv
            config_row: An optional row from `information_schema.tables_config` for the same mv
        Returns:
            A ReflectedMVState object.
        """
        ddl = mv_row.MATERIALIZED_VIEW_DEFINITION.strip()
        mv_fqn = utils.gen_simple_qualified_name(mv_row.TABLE_NAME, mv_row.TABLE_SCHEMA)
        # logger.debug("mv create ddl for %r: %s", mv_fqn, ddl)

        mv_name, schema = mv_row.TABLE_NAME, mv_row.TABLE_SCHEMA
        # 1. Parse the DDL to get properties that are only available there.
        #   Includes partition, refresh, and properties.
        try:
            parsed_state = self._parse_mv_ddl(mv_name, ddl, schema)
        except Exception as e:
            logger.warning(f"Failed to parse DDL for MV '{mv_row.TABLE_SCHEMA}.{mv_row.TABLE_NAME}', reflection may be incomplete: {e}")
            parsed_state = ReflectedMVState(table_name=mv_row.TABLE_NAME, definition=ddl)
        # logger.debug("partial parsed mv state. mv: %s, state: %s", mv_fqn, parsed_state)

        # 2. Augment/overwrite with more reliable info from other sources.
        parsed_state.table_options.update(self._parse_common_table_options(TableKind.MATERIALIZED_VIEW, table_row))

        if config_row:
            general_options = self._parse_general_table_options(mv_name, schema, table_config=config_row)
            logger.debug("parsed general table options for mv: %s, options: %s", mv_fqn, general_options)
            parsed_state.table_options.update(general_options)

        logger.debug("parsed mv state. mv: %s, state: %s", mv_fqn, parsed_state)
        return parsed_state

    def _parse_mv_ddl(
        self,
        mv_name: str,
        create_mv_ddl: str,
        schema: Optional[str] = None
    ) -> ReflectedMVState:
        """
        Parses the DDL from SHOW CREATE MATERIALIZED VIEW.
        This is the main entry point for MV reflection.
        """
        # Extract AS SELECT definition first
        mv_definition = None
        definition_match = self._MV_AS_DEFINITION_PATTERN.search(create_mv_ddl)
        if definition_match:
            mv_definition = definition_match.group(1).strip()
            clauses_str = create_mv_ddl[:definition_match.start()]
        else:
            raise SQLParseError(f"Could not find 'AS SELECT' in CREATE MATERIALIZED VIEW statement for {mv_name}", create_mv_ddl)

        state = ReflectedMVState(
            table_name=mv_name,
            definition=mv_definition,
        )

        partition_match = self._PARTITION_BY_PATTERN.search(clauses_str)
        if partition_match:
            state.table_options[TableInfoKeyWithPrefix.PARTITION_BY] = self.parse_partition_clause(partition_match.group(1).strip())

        # Use Lark parser for the refresh clause
        try:
            refresh_text_match = self._MV_REFRESH_PATTERN.search(clauses_str)
            if refresh_text_match:
                refresh_clause_str = refresh_text_match.group(0).strip()
                logger.debug("mv: %r, refresh_clause: %r", mv_name, refresh_clause_str)
                parsed_refresh = parse_mv_refresh_clause(refresh_clause_str)
                state.table_options[TableInfoKeyWithPrefix.REFRESH] = ReflectedRefreshInfo(
                    moment=parsed_refresh.get("refresh_moment"),
                    type=parsed_refresh.get("refresh_type")
                )
        except Exception as e:
            logger.warning(f"Failed to parse refresh clause for MV {mv_name}, falling back to regex: {e}")
            # Fallback to simple regex if lark parsing fails
            self._parse_mv_refresh_with_regex(clauses_str, state)

        # NOTE: currently, it uses properties from information_schema.tables_config, not from the DDL.
        # properties_match = self._MV_PROPERTIES_PATTERN.search(clauses_str)
        # if properties_match:
        #     # Use string instead of dictionary now.
        #     # state.mv_options.properties = self._parse_properties(properties_match.group(1))
        #     state.table_options[TableInfoKeyWithPrefix.PROPERTIES] = properties_match.group(1).strip()

        return state

    def _parse_mv_refresh_with_regex(self, ddl_part: str, state: ReflectedMVState):
        """Fallback refresh clause parser using regex."""
        refresh_match = re.search(r"REFRESH\s+(.+?)(?=\s*(?:PROPERTIES|AS))", ddl_part, re.IGNORECASE | re.DOTALL)
        if refresh_match:
            refresh_text = refresh_match.group(1).strip().upper()
            moment, type = None, None
            parts = refresh_text.split()
            if len(parts) > 0:
                if parts[0] in ("IMMEDIATE", "DEFERRED"):
                    moment = parts.pop(0)
                if parts:
                    type = " ".join(parts)
            state.table_options[TableInfoKeyWithPrefix.REFRESH] = ReflectedRefreshInfo(moment=moment, type=type)

    def _parse_properties(self, props_str: str) -> Dict[str, str]:
        """
        Parses the content of a PROPERTIES clause into a dictionary.
        This implementation uses regex to correctly handle commas and escaped quotes within property values.
        """
        # Regex to find all key-value pairs, respecting quotes.
        # It captures the key in group 1 and the value in group 2.
        # The value part `((?:\\"|[^"])*)` handles escaped quotes `\"` inside the value string.
        pattern = re.compile(r'"([^"]+)"\s*=\s*"((?:\\"|[^"])*)"')
        matches = pattern.findall(props_str)
        if matches:
            properties = {key: value.replace('\\"', '"') for key, value in matches}
        else:
            properties = {}
        return properties



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib.resources
import logging
from typing import Any, List, Tuple

from lark import Lark, Token, Transformer, v_args
from sqlalchemy.sql import sqltypes

from .. import datatype


logger = logging.getLogger(__name__)


class DataTypeTransformer(Transformer):
    """
    Use lark-parser to parse the data type string.
    And generate the corresponding SQLAlchemy type object.
    """

    def __init__(self, visit_tokens: bool = True) -> None:
        super().__init__(visit_tokens)
        import starrocks.dialect
        self.type_map = starrocks.dialect.StarRocksDialect.ischema_names.copy()
        # `int` is a keyword in lark grammar, but we want to map it to INTEGER
        self.type_map["int"] = self.type_map["integer"]

    def CNAME(self, t: Token) -> str:
        return t.value

    def NUMBER(self, n: Token) -> int:
        return int(n)

    @v_args(inline=True)
    def simple_type(self, type_name: str, args: List[int] = None, unsigned: Token = None) -> sqltypes.TypeEngine:
        type_name_lower = type_name.lower()
        # logger.debug(f"type_name_lower: {type_name_lower}, args: {args}, unsigned: {unsigned}")
        # There may be a bug of parsing the type and unsigned keyword together
        if args and (isinstance(args, Token) and args.value == "unsigned" or args == "unsigned"):
            args = None
            unsigned = True
        if type_name_lower == "bigint" and unsigned:
            return self.type_map["largeint"]

        # Convert TINYINT(1) to BOOLEAN during reflection
        # This ensures that TINYINT(1) from database is always reflected as BOOLEAN
        if type_name_lower == "tinyint" and args and len(args) == 1 and args[0] == 1:
            return self.type_map["boolean"]

        type_class = self.type_map.get(type_name_lower)
        # logger.debug(f"type_class: {type_class}")
        if type_class is None:
            # logger.error(f"Unsupported data type: {type_name}")
            raise TypeError(f"Unsupported data type: {type_name}")
        if args:
            return type_class(*args)
        return type_class

    def type_args(self, args: List[Any]) -> List[Any]:
        # logger.debug(f"args: {args}")
        return args

    @v_args(inline=True)
    def array_type(self, subtype: Any) -> datatype.ARRAY:
        # logger.debug(f"subtype: {subtype}")
        return datatype.ARRAY(subtype)

    @v_args(inline=True)
    def map_type(self, key_type: Any, value_type: Any) -> datatype.MAP:
        # logger.debug(f"key_type: {key_type}, value_type: {value_type}")
        return datatype.MAP(key_type, value_type)

    @v_args(inline=True)
    def struct_field(self, name: Any, field_type: Any) -> Tuple[Any, Any]:
        # logger.debug(f"struct field_name: {name}, field_type: {field_type}")
        return (name, field_type)

    def struct_type(self, fields: List[Tuple[Any, Any]]) -> datatype.STRUCT:
        # logger.debug(f"struct fields: {fields}")
        return datatype.STRUCT(*fields)

    @v_args(inline=True)
    def data_type(self, item: Any) -> Any:
        # logger.debug(f"data_type: {item}")
        return item


# For singleton pattern
_grammar_text = None
# For singleton pattern
_data_type_parser = None


def _get_grammar_text() -> str:
    global _grammar_text
    if _grammar_text is None:
        with importlib.resources.open_text("starrocks.drivers", "grammar.lark") as f:
            _grammar_text = f.read()
    return _grammar_text


def get_data_type_parser() -> Lark:
    """
    Returns a singleton instance of the data type Lark parser.
    """
    global _data_type_parser
    if _data_type_parser is None:
        grammar_text = _get_grammar_text()
        _data_type_parser = Lark(grammar_text, parser="lalr", start="data_type", transformer=DataTypeTransformer())
    return _data_type_parser


def parse_data_type(type_str: str) -> Any:
    """
    Parses a StarRocks type string and returns a SQLAlchemy type object.

    Args:
        type_str: The type string to parse.

    Returns:
        A SQLAlchemy type object.
    """
    return get_data_type_parser().parse(type_str)


# --- Materialized View Refresh Clause Parser ---
class _MVRefreshTransformer(Transformer):
    @v_args(inline=True)
    def refresh_clause(self, refresh_moment=None, refresh_type=None):
        # logger.debug(f"refresh_moment: {refresh_moment}, refresh_type: {refresh_type}")
        result = {"refresh_moment": refresh_moment, "refresh_type": refresh_type}
        if not refresh_type and refresh_moment \
                and refresh_moment.upper().strip() not in ("IMMEDIATE", "DEFERRED"):
            result["refresh_type"] = refresh_moment
            result["refresh_moment"] = None
        return result

    @v_args(inline=True)
    def refresh_moment(self, moment=None):
        # logger.debug(f"refresh_moment: {moment}")
        return str(moment).upper().strip() if moment else None

    @v_args(inline=True)
    def refresh_type(self, scheme=None):
        # logger.debug(f"refresh_type: {scheme}")
        return str(scheme).strip() if scheme else None

    @v_args(inline=True)
    def async_scheme(self, details=None):
        if details:
            return f"ASYNC {details}"
        return "ASYNC"

    def REFRESH(self, t: Token):
        return "REFRESH"

    def IMMEDIATE(self, t: Token):
        return "IMMEDIATE"

    def DEFERRED(self, t: Token):
        return "DEFERRED"

    def ASYNC(self, t: Token):
        return "ASYNC"

    def MANUAL(self, t: Token):
        return "MANUAL"

    def INCREMENTAL(self, t: Token):
        return "INCREMENTAL"

    def ASYNC_DETAILS(self, details):
        return str(details).strip()


_mv_refresh_parser = None


def _get_mv_refresh_parser() -> Lark:
    """Lazily loads and returns the Lark parser for MV refresh clauses."""
    global _mv_refresh_parser
    if _mv_refresh_parser is None:
        grammar_text = _get_grammar_text()
        _mv_refresh_parser = Lark(grammar_text, parser="lalr", start="refresh_clause", transformer=_MVRefreshTransformer())
    return _mv_refresh_parser


def parse_mv_refresh_clause(refresh_clause_str: str) -> dict:
    """Parses a full 'REFRESH ...' string and returns a dictionary with moment and scheme.
    Returns:
        A dictionary with keys "refresh_moment" and "refresh_type". The values are like:
        - "refresh_moment": "IMMEDIATE" | "DEFERRED" | None,
        - "refresh_type": "ASYNC" | "MANUAL" | "INCREMENTAL" | None,
    """
    return _get_mv_refresh_parser().parse(refresh_clause_str)



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sqlalchemy.testing.provision import configure_follower
from sqlalchemy.testing.provision import create_db
from sqlalchemy.testing.provision import drop_db
from sqlalchemy.testing.provision import temp_table_keyword_args

# from sqlalchemy import exc
# from sqlalchemy.testing.provision import generate_driver_url
#
#
# @generate_driver_url.for_db("starrocks")
# def generate_driver_url(url, driver, query_str):
#     backend = url.get_backend_name()
#
#     # NOTE: at the moment, tests are running mariadbconnector
#     # against both mariadb and mysql backends.   if we want this to be
#     # limited, do the decision making here to reject a "mysql+mariadbconnector"
#     # URL.  Optionally also re-enable the module level
#     # MySQLDialect_mariadbconnector.is_mysql flag as well, which must include
#     # a unit and/or functional test.
#
#     # all the Jenkins tests have been running mysqlclient Python library
#     # built against mariadb client drivers for years against all MySQL /
#     # MariaDB versions going back to MySQL 5.6, currently they can talk
#     # to MySQL databases without problems.
#
#     if backend == "starrocks":
#         dialect_cls = url.get_dialect()
#         if dialect_cls._is_mariadb_from_url(url):
#             backend = "mariadb"
#
#     new_url = url.set(
#         drivername="%s+%s" % (backend, driver)
#     ).update_query_string(query_str)
#
#     try:
#         new_url.get_dialect()
#     except exc.NoSuchModuleError:
#         return None
#     else:
#         return new_url
#

@create_db.for_db("starrocks")
def _mysql_create_db(cfg, eng, ident):
    with eng.begin() as conn:
        try:
            _starrocks_drop_db(cfg, conn, ident)
        except Exception:
            pass

    with eng.begin() as conn:
        conn.exec_driver_sql(
            "CREATE DATABASE %s" % ident
        )
        conn.exec_driver_sql(
            "CREATE DATABASE %s_test_schema" % ident
        )
        conn.exec_driver_sql(
            "CREATE DATABASE %s_test_schema_2" % ident
        )


@configure_follower.for_db("starrocks")
def _starrocks_configure_follower(config, ident):
    config.test_schema = "%s_test_schema" % ident
    config.test_schema_2 = "%s_test_schema_2" % ident


@drop_db.for_db("starrocks")
def _starrocks_drop_db(cfg, eng, ident):
    with eng.begin() as conn:
        conn.exec_driver_sql("DROP DATABASE %s_test_schema" % ident)
        conn.exec_driver_sql("DROP DATABASE %s_test_schema_2" % ident)
        conn.exec_driver_sql("DROP DATABASE %s" % ident)

@temp_table_keyword_args.for_db("starrocks")
def _starrocks_temp_table_keyword_args(cfg, eng):
    return {"prefixes": ["TEMPORARY"]}

# Uncomment to debug SQL Statements in tests
# from sqlalchemy.testing.provision import update_db_opts
# @update_db_opts.for_db("starrocks")
# def _starrocks_update_db_opts(db_url, db_opts):
#     db_opts["echo"] = True



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decimal
import logging

from sqlalchemy import (
    Integer,
    MetaData,
    String,
    Table,
    and_,
    cast,
    column,
    literal,
    schema,
    select,
    table,
    testing,
    text,
    type_coerce,
    union,
)
from sqlalchemy.engine import ObjectKind, ObjectScope
from sqlalchemy.sql.sqltypes import CHAR, Float
from sqlalchemy.testing import config, fixtures
from sqlalchemy.testing.assertions import AssertsCompiledSQL, eq_
from sqlalchemy.testing.schema import Column
from sqlalchemy.testing.suite import *  # noqa: F403, I001
from sqlalchemy.testing.suite import (
    ComponentReflectionTest as _ComponentReflectionTest,
    CTETest as _CTETest,
    CompositeKeyReflectionTest,
    EnumTest,
    FetchLimitOffsetTest as _FetchLimitOffsetTest,
    JSONTest as _JSONTest,
    LongNameBlowoutTest,
    NumericTest as _NumericTest,
    ServerSideCursorsTest as _ServerSideCursorsTest,
    StringTest as _StringTest,
)
from sqlalchemy.testing.util import mock

from starrocks.datatype import INTEGER
from test import test_utils


logger = logging.getLogger(__name__)


class SRAssertsCompiledSQLMixin:
    def assert_compile_normalized(self, clause, expected, **kw):
        """Compile and compare SQL using normalize_sql for format-independent comparison."""

        dialect = kw.get('dialect')
        if dialect is None:
            dialect = getattr(self, "__dialect__", None)
        if dialect is None:
            dialect = config.db.dialect
        # logger.debug(f"dialect: {dialect}")

        compiled = clause.compile(dialect=dialect, **kw.get('compile_kwargs', {}))
        actual_sql = str(compiled)

        normalized_actual = test_utils.normalize_sql(actual_sql)
        normalized_expected = test_utils.normalize_sql(expected)

        assert normalized_actual == normalized_expected, (
            f"\nActual (normalized):   {normalized_actual}\n"
            f"Expected (normalized): {normalized_expected}\n"
            f"Actual (raw):\n{actual_sql}"
        )


class CompileTest(fixtures.TestBase, AssertsCompiledSQL, SRAssertsCompiledSQLMixin):

    __only_on__ = "starrocks"

    def test_create_table_with_properties(self):
        m = MetaData()
        tbl = Table(
            'atable', m, Column("id", INTEGER),
            starrocks_properties=(
                ("storage_medium", "SSD"),
                ("storage_cooldown_time", "2015-06-04 00:00:00"),
                ("replication_num", "1"),
            ))
        self.assert_compile_normalized(
            schema.CreateTable(tbl),
            'CREATE TABLE atable (id INTEGER) PROPERTIES("storage_medium"="SSD", "storage_cooldown_time"="2015-06-04 00:00:00", "replication_num"="1")')

    def test_create_primary_key_table(self):
        m = MetaData()
        tbl = Table(
            'btable', m, Column("id", INTEGER, primary_key=True, autoincrement=False), # primary key must be specified autoincrement=False if not wanted to be auto generated
            starrocks_primary_key="id",
            starrocks_distributed_by="HASH(id)",
            starrocks_properties={"replication_num": "3"}
        )
        self.assert_compile_normalized(
            schema.CreateTable(tbl),
            'CREATE TABLE btable (id INTEGER NOT NULL)PRIMARY KEY(id) DISTRIBUTED BY HASH(id)PROPERTIES("replication_num"="3")')


class StarrocksMogrifyTest(fixtures.TablesTest, AssertsCompiledSQL):
    # Not strictly a dialect test, but allows me to test why mogrify is not working correctly
    def test_mogrify_query_with_parameters(self, connection):
        t = table('t1', column('c1'), column('c2'), column('test_param'))

        sel = select(
            cast(t.c.c1, CHAR(4000)),
            cast(t.c.c2, CHAR(4000)),
        ).where(
            and_(
                t.c.test_param == 'Y',
                ),
        )

        self.assert_compile(
            sel,
            result="SELECT CAST(t1.c1 AS CHAR(4000)) AS c1, CAST(t1.c2 AS CHAR(4000)) AS c2 FROM t1 WHERE t1.test_param = %(test_param_1)s",
            params={'test_param_1':'Y'},
            render_postcompile=True,
        )
        compiled = sel.compile(connection, compile_kwargs={"render_postcompile": True})
        mog = connection.engine.raw_connection().cursor().mogrify(str(compiled).replace('\n', ''), compiled.params)
        assert mog == "SELECT CAST(t1.c1 AS CHAR(4000)) AS c1, CAST(t1.c2 AS CHAR(4000)) AS c2 FROM t1 WHERE t1.test_param = 'Y'"

    # def test_select_nonrecursive_round_trip(self, connection):
    #     some_table = self.tables.some_table
    #
    #     cte = (
    #         select(some_table)
    #         .where(some_table.c.data.in_(["d2", "d3", "d4"]))
    #         .cte("some_cte")
    #     )
    #     result = connection.execute(
    #         select(cte.c.data).where(cte.c.data.in_(["d4", "d5"]))
    #     )
    #     eq_(result.fetchall(), [("d4",)])

class ComponentReflectionTest(_ComponentReflectionTest):
    # Updated to allow tests to run when Starrocks does not support default autoincrement on single primary key columns
    @classmethod
    def define_tables(cls, metadata):
        super().define_tables(metadata)
        for t_name, t in metadata.tables.items():
            if t.autoincrement_column is not None:
                t.autoincrement_column.autoincrement = True

    # Updated because Starrocks does not currently support column comments
    def exp_columns(
            self,
            schema=None,
            scope=ObjectScope.ANY,
            kind=ObjectKind.ANY,
            filter_names=None,
    ):
        def col(
                name, auto=False, default=mock.ANY, comment=None, nullable=True
        ):
            res = {
                "name": name,
                "autoincrement": auto,
                "type": mock.ANY,
                "default": default,
                "comment": comment if config.requirements.comment_reflection.enabled else '',
                "nullable": nullable,
            }
            if auto == "omit":
                res.pop("autoincrement")
            return res

        def pk(name, **kw):
            kw = {"auto": True, "default": mock.ANY, "nullable": False, **kw}
            return col(name, **kw)

        materialized = {
            (schema, "dingalings_v"): [
                col("dingaling_id", auto="omit", nullable=mock.ANY),
                col("address_id"),
                col("id_user"),
                col("data"),
            ]
        }
        views = {
            (schema, "email_addresses_v"): [
                col("address_id", auto="omit", nullable=mock.ANY),
                col("remote_user_id"),
                col("email_address"),
            ],
            (schema, "users_v"): [
                col("user_id", auto="omit", nullable=mock.ANY),
                col("test1", nullable=mock.ANY),
                col("test2", nullable=mock.ANY),
                col("parent_user_id"),
            ],
            (schema, "user_tmp_v"): [
                col("id", auto="omit", nullable=mock.ANY),
                col("name"),
                col("foo"),
            ],
        }
        self._resolve_views(views, materialized)
        tables = {
            (schema, "users"): [
                pk("user_id"),
                col("test1", nullable=False),
                col("test2", nullable=False),
                col("parent_user_id"),
            ],
            (schema, "dingalings"): [
                pk("dingaling_id"),
                col("address_id"),
                col("id_user"),
                col("data"),
            ],
            (schema, "email_addresses"): [
                pk("address_id"),
                col("remote_user_id"),
                col("email_address"),
            ],
            (schema, "comment_test"): [
                pk("id", comment="id comment"),
                col("data", comment="data % comment"),
                col(
                    "d2",
                    comment=r"""Comment types type speedily ' " \ '' Fun!""",
                ),
                col("d3", comment="Comment\nwith\rescapes"),
            ],
            (schema, "no_constraints"): [col("data")],
            (schema, "local_table"): [pk("id"), col("data"), col("remote_id")],
            (schema, "remote_table"): [pk("id"), col("local_id"), col("data")],
            (schema, "remote_table_2"): [pk("id"), col("data")],
            (schema, "noncol_idx_test_nopk"): [col("q")],
            (schema, "noncol_idx_test_pk"): [pk("id"), col("q")],
            (schema, self.temp_table_name()): [
                pk("id"),
                col("name"),
                col("foo"),
            ],
        }
        res = self._resolve_kind(kind, tables, views, materialized)
        res = self._resolve_names(schema, scope, filter_names, res)
        return res

class FetchLimitOffsetTest(_FetchLimitOffsetTest):

    # Fixed by adding order_by
    def test_limit_render_multiple_times(self, connection):
        table = self.tables.some_table
        stmt = select(table.c.id).order_by(table.c.id).limit(1).scalar_subquery()

        u = union(select(stmt), select(stmt)).subquery().select()

        self._assert_result(
            connection,
            u,
            [
                (1,),
            ],
        )

    @testing.skip('starrocks', 'cannot render offset without limit')
    @testing.requires.offset
    def test_simple_offset(self, connection):
        pass

    @testing.skip('starrocks', 'cannot render offset without limit')
    @testing.requires.offset
    def test_simple_offset_zero(self, connection):
        pass

    @testing.skip('starrocks', 'cannot render offset without limit')
    @testing.requires.bound_limit_offset
    def test_bound_offset(self, connection):
        pass


class NumericTest(_NumericTest,):
    # Changed because Starrocks does not support Float as first column

    @testing.fixture
    def do_numeric_test(self, metadata, connection):
        def run(type_, input_, output, filter_=None, check_scale=False):
            # Fix table so the first column is not float
            t = Table("t", metadata, Column("a", Integer), Column("x", type_))
            t.create(connection)
            connection.execute(t.insert(), [{"a": 1, "x": x} for x in input_])

            result = {row[0] for row in connection.execute(select(t.c.x))}
            output = set(output)
            if filter_:
                result = {filter_(x) for x in result}
                output = {filter_(x) for x in output}
            eq_(result, output)
            if check_scale:
                eq_([str(x) for x in result], [str(x) for x in output])

            connection.execute(t.delete())

            # test that this is actually a number!
            # note we have tiny scale here as we have tests with very
            # small scale Numeric types.  PostgreSQL will raise an error
            # if you use values outside the available scale.
            if type_.asdecimal:
                test_value = decimal.Decimal("2.9")
                add_value = decimal.Decimal("37.12")
            else:
                test_value = 2.9
                add_value = 37.12

            connection.execute(t.insert(), {"x": test_value})
            assert_we_are_a_number = connection.scalar(
                select(type_coerce(t.c.x + add_value, type_))
            )
            eq_(
                round(assert_we_are_a_number, 3),
                round(test_value + add_value, 3),
            )

        return run

    @testing.fixture
    def literal_round_trip(self, metadata, connection):
        """test literal rendering"""

        # for literal, we test the literal render in an INSERT
        # into a typed column.  we can then SELECT it back as its
        # official type; ideally we'd be able to use CAST here
        # but MySQL in particular can't CAST fully

        def run(
                type_,
                input_,
                output,
                filter_=None,
                compare=None,
                support_whereclause=True,
        ):
            if isinstance(type_, Float):
                t = Table("t", metadata, Column("a", Integer), Column("x", type_))
            else:
                t = Table("t", metadata, Column("x", type_))
            t.create(connection)

            for value in input_:
                ins = t.insert().values(
                    x=literal(value, type_, literal_execute=True)
                )
                connection.execute(ins)

            ins = t.insert().values(
                x=literal(None, type_, literal_execute=True)
            )
            connection.execute(ins)

            if support_whereclause and self.supports_whereclause:
                if compare:
                    stmt = select(t.c.x).where(
                        t.c.x
                        == literal(
                            compare,
                            type_,
                            literal_execute=True,
                        ),
                        t.c.x
                        == literal(
                            input_[0],
                            type_,
                            literal_execute=True,
                        ),
                        )
                else:
                    stmt = select(t.c.x).where(
                        t.c.x
                        == literal(
                            compare if compare is not None else input_[0],
                            type_,
                            literal_execute=True,
                        )
                    )
            else:
                stmt = select(t.c.x).where(t.c.x.is_not(None))

            rows = connection.execute(stmt).all()
            assert rows, "No rows returned"
            for row in rows:
                value = row[0]
                if filter_ is not None:
                    value = filter_(value)
                assert value in output

            stmt = select(t.c.x).where(t.c.x.is_(None))
            rows = connection.execute(stmt).all()
            eq_(rows, [(None,)])

        return run


class StringTest(_StringTest):
    # Fixed by adding order_by
    @testing.combinations(
        ("%B%", ["AB", "BC"]),
        ("A%C", ["AC"]),
        ("A%C%Z", []),
        argnames="expr, expected",
    )
    def test_dont_truncate_rightside(
            self, metadata, connection, expr, expected
    ):
        t = Table("t", metadata, Column("x", String(2)))
        t.create(connection)

        connection.execute(t.insert(), [{"x": "AB"}, {"x": "BC"}, {"x": "AC"}])

        eq_(
            connection.scalars(select(t.c.x).where(t.c.x.like(expr)).order_by(t.c.x)).all(),
            expected,
        )

class CTETest(_CTETest):
    @testing.requires.ctes_with_update_delete
    @testing.skip('starrocks', 'needs a primary column')
    def test_delete_scalar_subq_round_trip(self, connection):
        pass

    @testing.skip('starrocks', 'Does not support recursive CTE')
    def test_select_recursive_round_trip(self, connection):
        pass

class JSONTest(_JSONTest):
    # Override define_tables to specify autoincrement=True explicitly for test on Starrocks
    @classmethod
    def define_tables(cls, metadata):
        Table(
            "data_table",
            metadata,
            Column("id", Integer, primary_key=True, autoincrement=True),
            Column("name", String(30), nullable=False),
            Column("data", cls.datatype, nullable=False),
            Column("nulldata", cls.datatype(none_as_null=True)),
        )

    @testing.skip('starrocks', 'Seems to return "null", not sure why')
    def test_round_trip_json_null_as_json_null(self, connection):
        pass

    @testing.combinations(
        ("parameters",),
        ("multiparameters",),
        ("values",),
        argnames="insert_type",
    )
    @testing.skip("starrocks", 'Seems to return "null", not sure why')
    def test_round_trip_none_as_json_null(self, connection, insert_type):
        pass

    @testing.combinations(
        (True,),
        (False,),
        (None,),
        (15,),
        (0,),
        (-1,),
        (-1.0,),
        (15.052,),
        ("a string",),
        ("réve illé",),
        ("réve🐍 illé",),
    )
    @testing.skip("starrocks", 'Seems to return "null", not sure why')
    def test_single_element_round_trip(self, element):
        pass

class ServerSideCursorsTest(_ServerSideCursorsTest):

    @testing.combinations(
        ("global_string", True, lambda stringify: stringify("select 1"), True),
        (
                "global_text",
                True,
                lambda stringify: text(stringify("select 1")),
                True,
        ),
        ("global_expr", True, select(1), True),
        (
                "global_off_explicit",
                False,
                lambda stringify: text(stringify("select 1")),
                False,
        ),
        (
                "stmt_option",
                False,
                select(1).execution_options(stream_results=True),
                True,
        ),
        (
                "stmt_option_disabled",
                True,
                select(1).execution_options(stream_results=False),
                False,
        ),
        # Omit unsupported FOR UPDATE
        # ("for_update_expr", True, select(1).with_for_update(), True),
        # # TODO: need a real requirement for this, or dont use this test
        # (
        #     "for_update_string",
        #     True,
        #     lambda stringify: stringify("SELECT 1 FOR UPDATE"),
        #     True,
        #     testing.skip_if(["sqlite", "mssql"]),
        # ),
        (
                "text_no_ss",
                False,
                lambda stringify: text(stringify("select 42")),
                False,
        ),
        (
                "text_ss_option",
                False,
                lambda stringify: text(stringify("select 42")).execution_options(
                    stream_results=True
                ),
                True,
        ),
        id_="iaaa",
        argnames="engine_ss_arg, statement, cursor_ss_status",
    )
    def test_ss_cursor_status(
            self, engine_ss_arg, statement, cursor_ss_status
    ):
        engine = self._fixture(engine_ss_arg)
        with engine.begin() as conn:
            if callable(statement):
                statement = testing.resolve_lambda(
                    statement, stringify=self.stringify
                )

            if isinstance(statement, str):
                result = conn.exec_driver_sql(statement)
            else:
                result = conn.execute(statement)
            eq_(self._is_server_side(result.cursor), cursor_ss_status)
            result.close()

    @testing.skip("starrocks", 'Auto Increment seems to increment by 100000, not 1')
    def test_roundtrip_fetchall(self, metadata):
        pass

EnumTest.__requires__ = ("enums",)  # Fix Enum handling. Mysql has native ENUM type, but Starrocks has not
LongNameBlowoutTest.__requires__ = ("index_reflection",)  # This will do to make it skip for now, no multiple column index
CompositeKeyReflectionTest.__requires__ = ('primary_key_constraint_reflection',) # This will make it also skip the fixture which is not creating succesfully



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class SystemRunMode:
    """`run_mode` values defined by StarRocks"""
    SHARED_DATA = "shared_data"
    SHARED_NOTHING = "shared_nothing"


class TableEngine:
    OLAP = "OLAP"
    CLOUD_NATIVE = "CLOUD_NATIVE"


class TableDistribution:
    HASH = "HASH"
    RANDOM = "RANDOM"


class TableType:
    PRIMARY_KEY = "PRIMARY KEY"
    DUPLICATE_KEY = "DUPLICATE KEY"
    AGGREGATE_KEY = "AGGREGATE KEY"
    UNIQUE_KEY = "UNIQUE KEY"

    # For validation and mapping
    ALL_KEY_TYPES = {PRIMARY_KEY, DUPLICATE_KEY, AGGREGATE_KEY, UNIQUE_KEY}


class TableKey(TableType):
    """Alias name for TableType."""
    pass


class TableModel:
    """ Table type in information_schema.tables_config.TABLE_MODEL
    """
    DUP_KEYS = "DUP_KEYS"
    DUP_KEYS2 = "DUPLICATE_KEYS"
    AGG_KEYS = "AGG_KEYS"
    AGG_KEYS2 = "AGGREGATE_KEYS"
    PRI_KEYS = "PRI_KEYS"
    PRI_KEYS2 = "PRIMARY_KEYS"
    UNQ_KEYS = "UNQ_KEYS"
    UNQ_KEYS2 = "UNIQUE_KEYS"

    # useless
    # TO_TYPE_MAP = {
    #     DUP_KEYS: TableType.DUPLICATE_KEY,
    #     DUP_KEYS2: TableType.DUPLICATE_KEY,
    #     AGG_KEYS: TableType.AGGREGATE_KEY,
    #     AGG_KEYS2: TableType.AGGREGATE_KEY,
    #     PRI_KEYS: TableType.PRIMARY_KEY,
    #     PRI_KEYS2: TableType.PRIMARY_KEY,
    #     UNQ_KEYS: TableType.UNIQUE_KEY,
    #     UNQ_KEYS2: TableType.UNIQUE_KEY,
    # }


class PartitionType:
    RANGE = "RANGE"
    LIST = "LIST"
    EXPRESSION = "EXPRESSION"


class ColumnAggType:
    """Supported StarRocks aggregate functions for value columns.

    Use these constants in Column.info["starrocks_agg"].
    """

    # Key column type, rather than aggregate value type
    KEY = "KEY"

    # Core aggregate types
    SUM = "SUM"
    COUNT = "COUNT"
    MIN = "MIN"
    MAX = "MAX"

    # Specialized aggregate types
    HLL_UNION = "HLL_UNION"
    BITMAP_UNION = "BITMAP_UNION"
    REPLACE = "REPLACE"
    REPLACE_IF_NOT_NULL = "REPLACE_IF_NOT_NULL"

    # Allowed set for validation
    ALLOWED_ITEMS = {
        SUM,
        COUNT,
        MIN,
        MAX,
        HLL_UNION,
        BITMAP_UNION,
        REPLACE,
        REPLACE_IF_NOT_NULL,
    }


class ViewSecurityType:
    """Supported StarRocks view security types.

    Use these constants in View.info["security"].
    """

    EMPTY = ""
    NONE = "NONE"
    INVOKER = "INVOKER"
    DEFINER = "DEFINER"

    # Allowed set for validation
    ALLOWED_ITEMS = {
        EMPTY,
        NONE,
        INVOKER,
        DEFINER,
    }


class MVRefreshMoment:
    """Supported StarRocks materialized view refresh moments.
    """
    IMMEDIATE = "IMMEDIATE"
    DEFERRED = "DEFERRED"

    ALLOWED_ITEMS = {
        IMMEDIATE,
        DEFERRED,
    }


class MVRefreshType:
    """Supported StarRocks materialized view refresh schemes.
    """
    ASYNC = "ASYNC"
    MANUAL = "MANUAL"
    INCREMENTAL = "INCREMENTAL"

    ALLOWED_ITEMS = {
        ASYNC,
        MANUAL,
        INCREMENTAL,
    }



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Optional

from alembic.ddl.mysql import MySQLImpl
from sqlalchemy import Column, MetaData, Table

from starrocks import datatype
from starrocks.alembic import compare
from starrocks.datatype import BIGINT, VARCHAR


logger = logging.getLogger(__name__)


class StarRocksImpl(MySQLImpl):
    """Alembic DDL implementation for StarRocks."""

    __dialect__ = "starrocks"

    def version_table_impl(
        self,
        *,
        version_table: str,
        version_table_schema: Optional[str],
        version_table_pk: bool,  # ignored as StarRocks requires a primary key
        **kw,
    ) -> Table:
        version_table_kwargs = self.context_opts.get("version_table_kwargs", {}) if self.context_opts else {}
        if version_table_kwargs:
            logger.info(f"There are extra kwargs for version_table: {version_table_kwargs}")
        return Table(
            version_table,
            MetaData(),
            Column("id", BIGINT, autoincrement=True, primary_key=True),
            Column("version_num", VARCHAR(32), primary_key=False),
            schema=version_table_schema,
            starrocks_primary_key="id",
            **version_table_kwargs,
            **kw,
        )

    def compare_type(self, inspector_column: Column[Any], metadata_column: Column[Any]) -> bool:
        """
        Set StarRocks' specific type comparison logic for some special cases.

        For some special cases:
            - complex type comparison: ARRAY, MAP, STRUCT
            - simple type comparison:
                - meta.BOOLEAN equals to conn.TINYINT(1)
                - meta.STRING equals to conn.VARCHAR(65533)

        Args:
            inspector_column: The column from the inspector.
            metadata_column: The column from the metadata.

        Returns:
            True if the types are different, False if the types are the same.
        """
        inspector_type = inspector_column.type
        metadata_type = metadata_column.type

        # Handle complex type comparison.
        if isinstance(metadata_type, datatype.StructuredType):
            # If the inspector found a different base type, they are different.
            is_different = False
            if not isinstance(inspector_type, type(metadata_type)):
                is_different = True
            else:
                # Perform deep, recursive comparison.
                # Returns True if different, False if same.
                is_different = compare.compare_complex_type(self, inspector_type, metadata_type)
            if is_different:
                table: Optional[Table] = metadata_column.table
                table_info_msg = f" of table: {table.name}, schema: {table.schema}," if table is not None else ""
                logger.warning(f"Detected type change{table_info_msg} from inspector_type: {inspector_type!r} to metadata_type: {metadata_type!r}. "
                            f"But, StarRocks does not support schema change for complex type columns. "
                            f"You should check and change the metadata carefully to make sure there is no type difference for this column. "
                            )
            return is_different

        return compare.compare_simple_type(self, inspector_column, metadata_column)



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared constants for StarRocks dialect.

This module centralizes string keys and other constants that are used across
the StarRocks SQLAlchemy dialect to avoid typo-prone string literals and to
improve discoverability.
"""

from __future__ import annotations

from typing import Final


class TableConfigKey:
    """Keys found in information_schema.tables_config rows.

    Grouping related string literals as class-level constants helps avoid
    accidental typos and makes callsites self-documenting.
    """

    TABLE_ENGINE: Final[str] = "TABLE_ENGINE"
    TABLE_MODEL: Final[str] = "TABLE_MODEL"

    PARTITION_CLAUSE: Final[str] = "PARTITION_CLAUSE"  # Added, because only PARTITION_KEY in tables_config, which is not enough.
    # PARTITION_KEY: Final[str] = "PARTITION_KEY"

    DISTRIBUTE_TYPE: Final[str] = "DISTRIBUTE_TYPE"
    DISTRIBUTE_KEY: Final[str] = "DISTRIBUTE_KEY"
    DISTRIBUTE_BUCKET: Final[str] = "DISTRIBUTE_BUCKET"

    SORT_KEY: Final[str] = "SORT_KEY"
    PROPERTIES: Final[str] = "PROPERTIES"



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses
from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, TypedDict, Union

from sqlalchemy.engine.interfaces import ReflectedColumn

from starrocks.common.params import TableInfoKeyWithPrefix, TableKind


"""
Follow the mysql's ReflectedState, but with more specific types
It will be much cleaner and easier to use.
"""


def add_cached_str_clause(cls):
    """Add cached full string clause for a table attribute."""

    cls._cached_clause = None

    original_str = getattr(cls, "__str__", None)

    def __str__(self):
        if getattr(self, "_cached_clause", None) is None:
            self._cached_clause = original_str(self)
        return self._cached_clause

    cls.__str__ = __str__
    return cls


@dataclasses.dataclass
class ReflectedState:
    """Reflected information for a Table (base class). Similar to MySQL's implementation.
    comment is in table_options as the key `{dialect_name}_comment`.
    """
    table_name: Optional[str]
    columns: List[ReflectedColumn] = dataclasses.field(default_factory=list)
    table_options: Dict[str, Any] = dataclasses.field(default_factory=dict)
    keys: List[Union['ReflectedIndexInfo', 'ReflectedPKInfo', 'ReflectedUKInfo']] = dataclasses.field(default_factory=list)
    fk_constraints: List['ReflectedFKInfo'] = dataclasses.field(default_factory=list)
    ck_constraints: List['ReflectedCKInfo'] = dataclasses.field(default_factory=list)

    @property
    def name(self) -> Optional[str]:
        return self.table_name

    @name.setter
    def name(self, value: str) -> None:
        self.table_name = value

    @property
    def table_kind(self) -> str:
        return TableKind.TABLE

    @property
    def comment(self) -> Optional[str]:
        """Get the comment easily only, not able to write to it."""
        return self.table_options.get(TableInfoKeyWithPrefix.COMMENT)

    @property
    def partition_info(self) -> Optional['ReflectedPartitionInfo']:
        return self.table_options.get(TableInfoKeyWithPrefix.PARTITION_BY)

    @property
    def distribution_info(self) -> Optional['ReflectedDistributionInfo']:
        return self.table_options.get(TableInfoKeyWithPrefix.DISTRIBUTED_BY)

    @property
    def refresh_info(self) -> Optional['ReflectedRefreshInfo']:
        return self.table_options.get(TableInfoKeyWithPrefix.REFRESH)

    @property
    def order_by(self) -> Optional[str]:
        return self.table_options.get(TableInfoKeyWithPrefix.ORDER_BY)

    @property
    def properties(self) -> Optional[Union[str, Dict[str, str]]]:
        return self.table_options.get(TableInfoKeyWithPrefix.PROPERTIES)


@dataclasses.dataclass
class ReflectedViewState(ReflectedState):
    """Reflected information for a View"""
    definition: str = ''
    # security: Optional[str] = None # Moved to table_options

    # Views have no primary keys, foreign keys, or indexes
    keys: list = dataclasses.field(default_factory=list, init=False)
    fk_constraints: list = dataclasses.field(default_factory=list, init=False)

    @property
    def table_kind(self) -> str:
        return TableKind.VIEW

    @property
    def view_name(self) -> Optional[str]:
        return self.table_name

    @view_name.setter
    def view_name(self, value: str) -> None:
        self.table_name = value

    @property
    def security(self) -> Optional[str]:
        """Get the security type easily only, not able to write to it."""
        return self.table_options.get(TableInfoKeyWithPrefix.SECURITY)


@dataclasses.dataclass
class ReflectedMVState(ReflectedViewState):
    """Reflected information for a Materialized View"""

    @property
    def table_kind(self) -> str:
        return TableKind.MATERIALIZED_VIEW

    @property
    def mv_name(self) -> Optional[str]:
        return self.table_name

    @mv_name.setter
    def mv_name(self, value: str) -> None:
        self.table_name = value


@add_cached_str_clause
@dataclasses.dataclass
class ReflectedRefreshInfo:
    """Stores structured reflection information about a materialized view's refresh scheme."""
    moment: Optional[str] = None
    type: Optional[str] = None

    def __str__(self) -> str:
        parts = []
        if self.moment:
            parts.append(self.moment)
        if self.type:
            parts.append(self.type)
        return " ".join(parts)


class MySQLKeyType(Enum):
    PRIMARY = "PRIMARY"
    UNIQUE = "UNIQUE"
    FULLTEXT = "FULLTEXT"
    SPATIAL = "SPATIAL"


class ReflectedIndexInfo(TypedDict):
    """In ReflectedState.keys
    And, will be used to form ReflectedIndex
    """
    name: str
    type: MySQLKeyType
    parser: Optional[Any]  # Object?
    columns: List[Tuple[str, int, Any]]  # name, length, ...
    dialect_options: Dict[str, Any]


class ReflectedFKInfo(TypedDict):
    """In ReflectedState.fk_constraints
    And, will be used to form ReflectedForeignKeyConstraint
    """
    name: str
    table: NamedTuple[Optional[str], str]  # schema, name
    local: List[str]
    foreign: List[str]

    onupdate: bool
    ondelete: bool


class ReflectedPKInfo(TypedDict):
    """In ReflectedState.keys
    And, will be used to form ReflectedPrimaryKeyConstraint
    """
    type: MySQLKeyType
    columns: List[Tuple[str, Any, Any]]


class ReflectedUKInfo(TypedDict):
    """In ReflectedState.keys
    And, will be used to form ReflectedUniqueConstraint
    """
    name: str
    type: MySQLKeyType
    columns: List[Tuple[str, Any, Any]]


class ReflectedCKInfo(TypedDict):
    """In ReflectedState.ck_constraints
    And, will be used to form ReflectedCheckConstraint
    """
    name: str
    sqltext: str


@add_cached_str_clause
@dataclasses.dataclass(**dict(kw_only=True) if 'KW_ONLY' in dataclasses.__all__ else {})
class ReflectedTableKeyInfo:
    """
    Stores structed reflection information about a table' key/type.
    Such as `PRIMARY KEY (id, name)`, `UNIQUE KEY (id, name)`, `DUPLICATE KEY (id, name)`.

    Attributes:
        type: The key type string (e.g., 'PRIMARY KEY', 'UNIQUE KEY', 'DUPLICATE KEY').
        columns: The key columns as a list of strings or a single string (e.g., ['id', 'name'] or 'id, name').
    """
    type: str
    columns: Optional[Union[List[str], str]]

    def __str__(self) -> str:
        self.type = self.type.upper() if self.type else self.type
        if self.columns:
            self.columns = self.columns.strip()
        if isinstance(self.columns, list):
            return f"{self.type} ({', '.join(self.columns)})"
        return f"{self.type} ({self.columns})"

    def __repr__(self) -> str:
        return repr(str(self))


@add_cached_str_clause
@dataclasses.dataclass(**dict(kw_only=True) if 'KW_ONLY' in dataclasses.__all__ else {})
class ReflectedPartitionInfo:
    """
    Stores structured reflection information about a table's partitioning scheme.

    Attributes:
        type: The partitioning type (e.g., 'RANGE', 'LIST', 'EXPRESSION').
        partition_by: The whole partitioning expression string
            (e.g., 'RANGE(id, name)', 'date_trunc('day', dt)', 'id, col1, col2').
        pre_created_partitions: A string containing the full DDL for all
            pre-created partitions (e.g.,
            "(PARTITION p1 VALUES LESS THAN ('100'), PARTITION p2 VALUES LESS THAN ('200'))").
    """
    type: str
    partition_method: str  # includes the type and expressions
    pre_created_partitions: Optional[str] = None

    def __str__(self) -> str:
        self.type = self.type.upper() if self.type else self.type
        self.partition_method = self.partition_method.strip() if self.partition_method else self.partition_method
        if self.pre_created_partitions:
            return f"{self.partition_method} {self.pre_created_partitions}"
        return f"{self.partition_method}"

    def __repr__(self) -> str:
        return repr(str(self))


@add_cached_str_clause
@dataclasses.dataclass(**dict(kw_only=True) if 'KW_ONLY' in dataclasses.__all__ else {})
class ReflectedDistributionInfo:
    """Stores reflection information about a view."""
    type: Union[str, None]
    """The distribution type string like 'HASH' or 'RANDOM'."""
    columns: Optional[Union[List[str], str]] | None
    """The distribution columns string like 'id' or 'id, name'."""
    distribution_method: Union[str, None]
    """The distribution method string like 'HASH(id)' or 'RANDOM' without BUCKETS.
    It will be used first if it's not None."""
    buckets: Union[int, None]
    """The buckets count."""

    def __str__(self) -> str:
        """Convert to string representation of distribution option."""
        buckets_str = f' BUCKETS {self.buckets}' if self.buckets and str(self.buckets) != "0" else ""
        if not self.distribution_method:
            distribution_cols = ', '.join(self.columns) if isinstance(self.columns, list) else self.columns
            distribution_cols_str = f'({distribution_cols})' if distribution_cols else ""
            self.distribution_method = f'{self.type}{distribution_cols_str}'
        return f'{self.distribution_method}{buckets_str}'

    def __repr__(self) -> str:
        return repr(str(self))



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sqlalchemy.testing import fixtures
from sqlalchemy.testing.assertions import AssertsCompiledSQL
from sqlalchemy import (
    Table,
    Column,
    Integer,
    func,
    MetaData,
    literal_column,
)

from starrocks.sql.dml import (
    InsertIntoFiles,
    FilesTarget,
    FilesTargetOptions,
    InsertFromFiles,
    FilesSource,
    # FilesSourceOptions,
    CSVFormat,
    ParquetFormat,
    GoogleCloudStorage,
    Compression,
)


class CompileStarrocksInsertIntoFilesTest(fixtures.TestBase, AssertsCompiledSQL):

    __only_on__ = "starrocks"

    def test_insert_into_files(self):
        m = MetaData()
        tbl = Table(
            "atable",
            m,
            Column("id", Integer),
            schema="test_schema",
        )
        insert_into_files = InsertIntoFiles(
            target=FilesTarget(
                storage=GoogleCloudStorage(
                    uri='gs://starrocks/atable',
                    service_account_email='x@y.z',
                    service_account_private_key_id='mykey',
                    service_account_private_key='some_private_key',
                ),
                format=CSVFormat(
                    column_separator=',',
                    line_delimiter='\n',
                    enclose='"',
                    skip_header=1,
                ),
                options=FilesTargetOptions(
                    single=True,
                )
            ),
            from_=tbl.select(),
        )

        self.assert_compile(
            insert_into_files,
            (
                "INSERT INTO FILES("
                "'path' = 'gs://starrocks/atable',"
                "'gcp.gcs.service_account_email' = 'x@y.z',"
                "'gcp.gcs.service_account_private_key_id' = 'mykey',"
                "'gcp.gcs.service_account_private_key' = 'some_private_key',"
                "'format' = 'csv',"
                "'csv.line_delimiter' = '\\n',"
                "'csv.column_separator' = ',',"
                "'csv.enclose' = '\"',"
                "'csv.skip_header' = '1',"
                "'single' = 'true'"
                ")"
                " SELECT test_schema.atable.id FROM test_schema.atable"
            ),
        )


class CompileStarrocksInsertFromFilesTest(fixtures.TestBase, AssertsCompiledSQL):

    __only_on__ = "starrocks"

    def test_insert_from_files_csv(self):
        m = MetaData()
        tbl = Table(
            "atable",
            m,
            Column("id", Integer),
            schema="test_schema",
        )
        insert_from_files = InsertFromFiles(
            target=tbl,
            from_=FilesSource(
                storage=GoogleCloudStorage(
                    uri='gs://starrocks/atable',
                    service_account_email='x@y.z',
                    service_account_private_key_id='mykey',
                    service_account_private_key='some_private_key',
                ),
                format=CSVFormat(
                    column_separator=',',
                    row_delimiter='\n',
                    enclose='"',
                ),
                # options=FilesSourceOptions(
                # )
            ),
        )

        self.assert_compile(
            insert_from_files,
            (
                "INSERT INTO test_schema.atable "
                "SELECT * FROM FILES("
                "'path' = 'gs://starrocks/atable',"
                "'gcp.gcs.service_account_email' = 'x@y.z',"
                "'gcp.gcs.service_account_private_key_id' = 'mykey',"
                "'gcp.gcs.service_account_private_key' = 'some_private_key',"
                "'format' = 'csv',"
                "'csv.row_delimiter' = '\\n',"
                "'csv.column_separator' = ',',"
                "'csv.enclose' = '\"'"
                ")"
            ),
        )

    def test_insert_from_files_parquet(self):
        m = MetaData()
        tbl = Table(
            "atable",
            m,
            Column("id", Integer),
            schema="test_schema",
        )
        insert_from_files = InsertFromFiles(
            target=tbl,
            from_=FilesSource(
                storage=GoogleCloudStorage(
                    uri='gs://starrocks/atable.parquet',
                    service_account_email='x@y.z',
                    service_account_private_key_id='mykey',
                    service_account_private_key='some_private_key',
                ),
                format=ParquetFormat(
                    compression=Compression.SNAPPY
                ),
            ),
        )

        self.assert_compile(
            insert_from_files,
            (
                "INSERT INTO test_schema.atable "
                "SELECT * FROM FILES("
                "'path' = 'gs://starrocks/atable.parquet',"
                "'gcp.gcs.service_account_email' = 'x@y.z',"
                "'gcp.gcs.service_account_private_key_id' = 'mykey',"
                "'gcp.gcs.service_account_private_key' = 'some_private_key',"
                "'format' = 'parquet',"
                "'compression' = 'snappy'"
                ")"
            ),
        )

    def test_insert_from_files_column_str(self):
        m = MetaData()
        tbl = Table(
            "atable",
            m,
            Column("id", Integer),
            schema="test_schema",
        )
        insert_from_files = InsertFromFiles(
            target=tbl,
            from_=FilesSource(
                storage=GoogleCloudStorage(
                    uri='gs://starrocks/atable.parquet',
                    service_account_email='x@y.z',
                    service_account_private_key_id='mykey',
                    service_account_private_key='some_private_key',
                ),
                format=ParquetFormat(
                    compression=Compression.SNAPPY
                ),
            ),
            columns='$1, $2, $3'
        )

        self.assert_compile(
            insert_from_files,
            (
                "INSERT INTO test_schema.atable "
                "SELECT $1, $2, $3"
                " FROM FILES("
                "'path' = 'gs://starrocks/atable.parquet',"
                "'gcp.gcs.service_account_email' = 'x@y.z',"
                "'gcp.gcs.service_account_private_key_id' = 'mykey',"
                "'gcp.gcs.service_account_private_key' = 'some_private_key',"
                "'format' = 'parquet',"
                "'compression' = 'snappy'"
                ")"
            ),
        )

    def test_insert_from_files_column_expr(self):
        m = MetaData()
        tbl = Table(
            "atable",
            m,
            Column("id", Integer),
            schema="test_schema",
        )
        insert_from_files = InsertFromFiles(
            target=tbl,
            from_=FilesSource(
                storage=GoogleCloudStorage(
                    uri='gs://starrocks/atable.parquet',
                    service_account_email='x@y.z',
                    service_account_private_key_id='mykey',
                    service_account_private_key='some_private_key',
                ),
                format=ParquetFormat(
                    compression=Compression.SNAPPY
                ),
            ),
            columns=[func.IF(literal_column("$1") == "xyz", "NULL", "NOTNULL")]
        )

        self.assert_compile(
            insert_from_files,
            (
                "INSERT INTO test_schema.atable "
                "SELECT IF($1 = %(1_1)s, %(IF_1)s, %(IF_2)s)"
                " FROM FILES("
                "'path' = 'gs://starrocks/atable.parquet',"
                "'gcp.gcs.service_account_email' = 'x@y.z',"
                "'gcp.gcs.service_account_private_key_id' = 'mykey',"
                "'gcp.gcs.service_account_private_key' = 'some_private_key',"
                "'format' = 'parquet',"
                "'compression' = 'snappy'"
                ")"
            ),
            checkparams={"1_1": "xyz", "IF_1": "NULL", "IF_2": "NOTNULL"},
        )




# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import Dict, List, Optional, Union

from alembic.operations import Operations, ops
from sqlalchemy import MetaData, Table

from starrocks.common.params import (
    SRKwargsPrefix,
    TableObjectInfoKey,
)
from starrocks.sql.schema import MaterializedView, View, extract_view_columns


logger = logging.getLogger(__name__)


def _extract_starrocks_dialect_kwargs(kwargs: dict) -> dict:
    """
    Return only StarRocks dialect kwargs that start with the SRKwargsPrefix (e.g., 'starrocks_'),
    and normalize the suffix (after the prefix) to lowercase.

    Notes:
    - The 'starrocks_' prefix is required to be lowercase and matched case-sensitively.
    - The attribute part after the prefix is normalized to lowercase, so both
      'starrocks_PARTITION_BY' and 'starrocks_partition_by' are accepted and stored as
      'starrocks_partition_by'.
    """
    return {k.lower(): v for k, v in kwargs.items() if k.startswith(SRKwargsPrefix)}


def _columns_dicts_to_column_objects(columns: Union[List[Dict], None]) -> List:
    """
    Helper function to convert column dicts to Column objects.

    This is used by both CreateViewOp and CreateMaterializedViewOp when
    converting Op to View/MaterializedView objects for toimpl execution.

    Args:
        columns: List of column dicts with 'name' and optional 'comment' keys

    Returns:
        List of Column objects

    Example:
        >>> columns = [{'name': 'id', 'comment': 'ID'}, {'name': 'name'}]
        >>> _columns_dicts_to_column_objects(columns)
        [Column('id', STRING(), comment='ID'), Column('name', STRING())]
    """
    from sqlalchemy import Column

    from starrocks.datatype import STRING

    if not columns:
        return []

    result = []
    for col_dict in columns:
        col = Column(col_dict['name'], STRING())
        if col_dict.get('comment'):
            col.comment = col_dict['comment']
        result.append(col)
    return result


@Operations.register_operation("alter_view")
class AlterViewOp(ops.MigrateOperation):
    """Represent an ALTER VIEW operation."""
    def __init__(
        self,
        view_name: str,
        definition: Optional[str] = None,
        schema: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        comment: Optional[str] = None,
        security: Optional[str] = None,
        existing_definition: Optional[str] = None,
        existing_columns: Union[List[Dict], None] = None,
        existing_comment: Optional[str] = None,
        existing_security: Optional[str] = None,
    ):
        """
        Definition usually should not be None. But for future, we may want to support Alter View by only comment or security,
        so we keep the definition as Noneable.
        """
        self.view_name = view_name
        self.definition = definition
        self.schema = schema
        self.columns = columns
        self.comment = comment
        self.security = security
        self.existing_definition = existing_definition
        self.existing_columns = existing_columns
        self.existing_comment = existing_comment
        self.existing_security = existing_security

    def to_view(self, metadata: Optional[MetaData] = None) -> "View":
        return View(
            self.view_name,
            MetaData(),
            definition=self.definition or '<not_used_definition>',
            schema=self.schema,
            columns=self.columns,
            comment=self.comment,
            starrocks_security=self.security,
        )

    @classmethod
    def alter_view(
        cls,
        operations: Operations,
        view_name: str,
        definition: Optional[str] = None,
        schema: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        comment: Optional[str] = None,
        security: Optional[str] = None,
        existing_definition: Optional[str] = None,
        existing_columns: Union[List[Dict], None] = None,
        existing_comment: Optional[str] = None,
        existing_security: Optional[str] = None,
    ):
        """Invoke an ALTER VIEW operation."""
        op = cls(
            view_name,
            definition,
            schema=schema,
            columns=columns,
            comment=comment,
            security=security,
            existing_definition=existing_definition,
            existing_columns=existing_columns,
            existing_comment=existing_comment,
            existing_security=existing_security,
        )
        return operations.invoke(op)

    def reverse(self) -> "AlterViewOp":
        # Reversing an ALTER is another ALTER, using the stored "reverse" attributes.
        logger.debug("reverse AlterViewOp for %s", self.view_name)

        return AlterViewOp(
            self.view_name,
            definition=self.existing_definition,
            schema=self.schema,
            columns=self.existing_columns,
            comment=self.existing_comment,
            security=self.existing_security,
            existing_definition=self.definition,
            existing_columns=self.columns,
            existing_comment=self.comment,
            existing_security=self.security,
        )

    def __str__(self) -> str:
        """String representation for debugging."""
        return (
            f"AlterViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"definition=({self.definition!r}), columns={self.columns!r}, comment={self.comment!r}, "
            f"security={self.security}, "
            f"existing_definition=({self.existing_definition}), "
            f"existing_columns={self.existing_columns!r}, "
            f"existing_comment={self.existing_comment!r}, "
            f"existing_security={self.existing_security})"
        )


@Operations.register_operation("create_view")
class CreateViewOp(ops.MigrateOperation):
    def __init__(
        self,
        view_name: str,
        definition: str,
        schema: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        comment: Optional[str] = None,
        or_replace: bool = False,
        if_not_exists: bool = False,
        **kwargs,
    ) -> None:
        self.view_name = view_name
        self.definition = definition
        self.schema = schema
        self.columns = columns
        self.comment = comment
        self.or_replace = or_replace
        self.if_not_exists = if_not_exists
        self.kwargs = _extract_starrocks_dialect_kwargs(kwargs)

    def to_view(self, metadata: Optional[MetaData] = None) -> "View":
        return View(
            self.view_name,
            MetaData(),
            definition=self.definition,
            schema=self.schema,
            columns=self.columns,
            comment=self.comment,
            **self.kwargs,
        )

    @classmethod
    def from_view(cls, view: Table) -> "CreateViewOp":
        """Create Op from a View object (which is a Table)."""
        return cls(
            view.name,
            definition=view.info.get(TableObjectInfoKey.DEFINITION),
            schema=view.schema,
            columns=extract_view_columns(view),
            comment=view.comment,
            **view.dialect_kwargs,
        )

    @classmethod
    def create_view(
        cls,
        operations: Operations,
        view_name: str,
        definition: str,
        schema: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        comment: Optional[str] = None,
        or_replace: bool = False,
        if_not_exists: bool = False,
        **kwargs,
    ) -> None:
        op = cls(
            view_name,
            definition,
            schema=schema,
            columns=columns,
            comment=comment,
            or_replace=or_replace,
            if_not_exists=if_not_exists,
            **kwargs,
        )
        return operations.invoke(op)

    def reverse(self) -> "DropViewOp":
        # logger.debug("reverse CreateViewOp for %s", self.view_name)
        return DropViewOp(
            self.view_name,
            schema=self.schema,
            **self.kwargs,
        )

    def __str__(self) -> str:
        return (
            f"CreateViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"definition=({self.definition!r}), columns={self.columns!r}, comment={self.comment!r}, "
            f"or_replace={self.or_replace}, "
            f"if_not_exists={self.if_not_exists}, kwargs=({self.kwargs!r}), "
            f"kwargs={self.kwargs!r}"
        )


@Operations.register_operation("drop_view")
class DropViewOp(ops.MigrateOperation):
    def __init__(
        self,
        view_name: str,
        schema: Optional[str] = None,
        if_exists: bool = False,
        existing_definition: Optional[str] = None,
        existing_columns: Optional[List[Dict]] = None,
        existing_comment: Optional[str] = None,
        **kwargs,
    ):
        self.view_name = view_name
        self.schema = schema
        self.if_exists = if_exists
        self.existing_definition = existing_definition
        self.existing_columns = existing_columns
        self.existing_comment = existing_comment
        self.kwargs = _extract_starrocks_dialect_kwargs(kwargs)

    def to_view(self, metadata: Optional[MetaData] = None) -> "View":
        """Create a View object for DROP VIEW operation.

        This is used by toimpl to construct the DROP VIEW DDL statement.
        For DROP VIEW, we only need the view name and schema, not the full definition.
        """
        return View(
            self.view_name,
            MetaData(),
            definition='<not_used_definition>',  # Empty definition for DROP VIEW
            schema=self.schema,
            **self.kwargs,
        )

    @classmethod
    def from_view(cls, view: Table) -> "DropViewOp":
        """Create DropViewOp from a View object (which is a Table)."""
        return cls(
            view.name,
            schema=view.schema,
            existing_definition=view.info.get(TableObjectInfoKey.DEFINITION),
            existing_columns=extract_view_columns(view),
            existing_comment=view.comment,
            **view.dialect_kwargs,
        )

    @classmethod
    def drop_view(
        cls,
        operations: Operations,
        view_name: str,
        schema: Optional[str] = None,
        if_exists: bool = False,
        **kwargs,
    ):
        op = cls(view_name, schema=schema, if_exists=if_exists, **kwargs)
        return operations.invoke(op)

    def reverse(self) -> "CreateViewOp":
        if self.existing_definition is None:
            raise NotImplementedError("Cannot reverse a DropViewOp without the view's definition.")
        op = CreateViewOp(
            self.view_name,
            definition=self.existing_definition,
            schema=self.schema,
            columns=self.existing_columns,
            comment=self.existing_comment,
            **self.kwargs,
        )
        # logger.debug("reverse DropViewOp for %s, with op: (%s)", self.view_name, op)
        return op

    def __str__(self) -> str:
        return (
            f"DropViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"if_exists={self.if_exists}, existing_definition=({self.existing_definition!r}), "
            f"existing_comment=({self.existing_comment!r}), "
            f"existing_columns={self.existing_columns!r}), "
            f"kwargs=({self.kwargs!r})"
        )


@Operations.register_operation("alter_materialized_view")
class AlterMaterializedViewOp(ops.MigrateOperation):
    """
    Alter a materialized view (mutable attributes only).

    Based on StarRocks ALTER MATERIALIZED VIEW documentation:
    https://docs.starrocks.io/docs/sql-reference/sql-statements/materialized_view/ALTER_MATERIALIZED_VIEW/

    Only supports altering mutable attributes:
    - refresh: ALTER MATERIALIZED VIEW ... REFRESH <new_scheme>
    - properties: ALTER MATERIALIZED VIEW ... SET ("<key>" = "<value>")

    Immutable attributes (definition, partition_by, distributed_by, order_by, comment, columns)
    cannot be altered via ALTER MATERIALIZED VIEW and require DROP + CREATE.

    Currently, we make an Alter including all the mutable attributes, which is different with AlterTableXxxOp.
    """

    def __init__(
        self,
        view_name: str,
        schema: Optional[str] = None,
        # Mutable attributes only
        refresh: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        # Reverse values for downgrade
        existing_refresh: Optional[str] = None,
        existing_properties: Optional[Dict[str, str]] = None,
    ) -> None:
        self.view_name = view_name
        self.schema = schema
        self.refresh = refresh
        self.properties = properties
        self.existing_refresh = existing_refresh
        self.existing_properties = existing_properties

    @classmethod
    def alter_materialized_view(
        cls,
        operations: Operations,
        view_name: str,
        schema: Optional[str] = None,
        refresh: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        existing_refresh: Optional[str] = None,
        existing_properties: Optional[Dict[str, str]] = None,
    ):
        """Invoke an ALTER MATERIALIZED VIEW operation."""
        op = cls(
            view_name,
            schema=schema,
            refresh=refresh,
            properties=properties,
            existing_refresh=existing_refresh,
            existing_properties=existing_properties,
        )
        return operations.invoke(op)

    def reverse(self) -> "AlterMaterializedViewOp":
        return AlterMaterializedViewOp(
            self.view_name,
            schema=self.schema,
            refresh=self.existing_refresh,
            properties=self.existing_properties,
            existing_refresh=self.refresh,
            existing_properties=self.properties,
        )

    def __str__(self) -> str:
        """String representation for debugging."""
        return (
            f"AlterMaterializedViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"refresh={self.refresh!r}, "
            f"properties={self.properties!r}, "
            f"existing_refresh={self.existing_refresh!r}, "
            f"existing_properties={self.existing_properties!r})"
        )


@Operations.register_operation("create_materialized_view")
class CreateMaterializedViewOp(CreateViewOp):
    """
    Create a materialized view.

    Inherits from CreateViewOp to reuse view_name, definition, schema, comment, columns.
    Adds dialect-specific parameters via kwargs, e.g., starrocks_partition_by.
    """

    def __init__(
        self,
        view_name: str,
        definition: str,
        schema: Optional[str] = None,
        comment: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        or_replace: bool = False,
        if_not_exists: bool = False,
        **kwargs,
    ):
        super().__init__(
            view_name,
            definition,
            schema=schema,
            comment=comment,
            columns=columns,
            or_replace=or_replace,
            if_not_exists=if_not_exists,
            **kwargs,  # Processed by CreateViewOp.__init__ to only keep starrocks_*
        )

    def to_mv(self, metadata: Optional[MetaData] = None) -> "MaterializedView":
        """Convert Op to MaterializedView object (for toimpl)."""
        # Convert columns dict to Column objects
        columns = _columns_dicts_to_column_objects(self.columns)

        return MaterializedView(
            self.view_name,
            metadata or MetaData(),
            *columns,
            definition=self.definition,
            schema=self.schema,
            comment=self.comment,
            **self.kwargs,
        )

    # Keep backward compatibility alias
    def to_materialized_view(self, metadata: Optional[MetaData] = None) -> "MaterializedView":
        """Backward compatibility alias for to_mv()."""
        return self.to_mv(metadata)

    @classmethod
    def from_mv(cls, mv: Table) -> "CreateMaterializedViewOp":
        """Create Op from MaterializedView Table object."""
        columns = extract_view_columns(mv)

        return cls(
            view_name=mv.name,
            definition=mv.info.get(TableObjectInfoKey.DEFINITION),
            schema=mv.schema,
            comment=mv.comment,
            columns=columns,
            **mv.dialect_kwargs,
        )

    # Keep backward compatibility alias
    @classmethod
    def from_materialized_view(cls, mv: Table) -> "CreateMaterializedViewOp":
        """Backward compatibility alias for from_mv()."""
        return cls.from_mv(mv)

    @classmethod
    def create_materialized_view(
        cls,
        operations: Operations,
        view_name: str,
        definition: str,
        schema: Optional[str] = None,
        comment: Optional[str] = None,
        columns: Union[List[Dict], None] = None,
        or_replace: bool = False,
        if_not_exists: bool = False,
        **kwargs,
    ):
        """Invoke a CREATE MATERIALIZED VIEW operation."""
        op = cls(
            view_name,
            definition,
            schema=schema,
            comment=comment,
            columns=columns,
            or_replace=or_replace,
            if_not_exists=if_not_exists,
            **kwargs,
        )
        return operations.invoke(op)

    def reverse(self) -> "DropMaterializedViewOp":
        return DropMaterializedViewOp(
            self.view_name,
            schema=self.schema,
        )

    def __str__(self) -> str:
        """String representation for debugging."""
        return (
            f"CreateMaterializedViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"definition={self.definition!r}, comment={self.comment!r}, columns={self.columns!r}, "
            f"or_replace={self.or_replace}, if_not_exists={self.if_not_exists}, "
            f"kwargs={self.kwargs!r})"
        )


@Operations.register_operation("drop_materialized_view")
class DropMaterializedViewOp(DropViewOp):
    """Drop a materialized view operation."""

    def __init__(
        self,
        view_name: str,
        schema: Optional[str] = None,
        if_exists: bool = False,
        existing_definition: Optional[str] = None,
        existing_columns: Optional[List[Dict]] = None,
        existing_comment: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        kwargs are already processed by DropViewOp.__init__ to only keep starrocks_*.
        """
        super().__init__(
            view_name,
            schema=schema,
            if_exists=if_exists,
            existing_definition=existing_definition,
            existing_columns=existing_columns,
            existing_comment=existing_comment,
            **kwargs,
        )

    @classmethod
    def from_mv(cls, mv: Table) -> "DropMaterializedViewOp":
        """Create DropMaterializedViewOp from a MaterializedView object."""
        return cls(
            mv.name,
            schema=mv.schema,
            existing_definition=mv.info.get(TableObjectInfoKey.DEFINITION),
            existing_columns=extract_view_columns(mv),
            existing_comment=mv.comment,
            **mv.dialect_kwargs,
        )

    @classmethod
    def from_materialized_view(cls, mv: Table) -> "DropMaterializedViewOp":
        return cls.from_mv(mv)

    @classmethod
    def drop_materialized_view(
        cls,
        operations: Operations,
        view_name: str,
        schema: Optional[str] = None,
        if_exists: bool = False,
        **kwargs,
    ):
        """Invoke a DROP MATERIALIZED VIEW operation."""
        op = cls(view_name, schema=schema, if_exists=if_exists, **kwargs)
        return operations.invoke(op)

    def reverse(self) -> "CreateMaterializedViewOp":
        if self.existing_definition is None:
            raise NotImplementedError("Cannot reverse a DropMaterializedViewOp without the view's definition.")
        return CreateMaterializedViewOp(
            self.view_name,
            definition=self.existing_definition,
            schema=self.schema,
            comment=self.existing_comment,
            columns=self.existing_columns,
            **self.kwargs,
        )

    def __str__(self) -> str:
        """String representation for debugging."""
        return (
            f"DropMaterializedViewOp(view_name={self.view_name!r}, schema={self.schema!r}, "
            f"if_exists={self.if_exists}, existing_definition={self.existing_definition!r}, "
            f"kwargs={self.kwargs!r})"
        )


# Operation classes ordered according to StarRocks grammar:
# engine → key → (comment) → partition → distribution → order by → properties
@Operations.register_operation("alter_table_engine")
class AlterTableEngineOp(ops.AlterTableOp):
    """Represent an ALTER TABLE ENGINE operation for StarRocks."""

    def __init__(
        self,
        table_name: str,
        engine: str,
        schema: Optional[str] = None,
        existing_engine: Optional[str] = None,
    ):
        super().__init__(table_name, schema=schema)
        self.engine = engine
        self.existing_engine = existing_engine

    @classmethod
    def alter_table_engine(
        cls,
        operations: Operations,
        table_name: str,
        engine: str,
        schema: Optional[str] = None,
        existing_engine: Optional[str] = None,
    ):
        """Invoke an ALTER TABLE ENGINE operation for StarRocks."""
        op = cls(table_name, engine, schema=schema, existing_engine=existing_engine)
        return operations.invoke(op)

    def reverse(self) -> AlterTableEngineOp:
        if self.existing_engine is None:
            raise NotImplementedError("Cannot reverse AlterTableEngineOp without existing_engine")
        return AlterTableEngineOp(
            table_name=self.table_name,
            engine=self.existing_engine,
            schema=self.schema,
            existing_engine=self.engine,
        )

    def __str__(self) -> str:
        return (f"AlterTableEngineOp(table_name={self.table_name!r}, "
               f"engine={self.engine!r}, schema={self.schema!r}, "
               f"existing_engine={self.existing_engine!r})")

@Operations.register_operation("alter_table_key")
class AlterTableKeyOp(ops.AlterTableOp):
    """Represent an ALTER TABLE KEY operation for StarRocks."""

    def __init__(
        self,
        table_name: str,
        key_type: str,
        key_columns: str,
        schema: Optional[str] = None,
        existing_key_type: Optional[str] = None,
        existing_key_columns: Optional[str] = None,
    ):
        super().__init__(table_name, schema=schema)
        self.key_type = key_type
        self.key_columns = key_columns
        self.existing_key_type = existing_key_type
        self.existing_key_columns = existing_key_columns

    @classmethod
    def alter_table_key(
        cls,
        operations: Operations,
        table_name: str,
        key_type: str,
        key_columns: str,
        schema: Optional[str] = None,
        existing_key_type: Optional[str] = None,
        existing_key_columns: Optional[str] = None,
    ):
        """Invoke an ALTER TABLE KEY operation for StarRocks."""
        op = cls(
            table_name,
            key_type,
            key_columns,
            schema=schema,
            existing_key_type=existing_key_type,
            existing_key_columns=existing_key_columns,
        )
        return operations.invoke(op)

    def reverse(self) -> "AlterTableKeyOp":
        if self.existing_key_type is None or self.existing_key_columns is None:
            raise NotImplementedError("Cannot reverse AlterTableKeyOp without existing_key_type and existing_key_columns")
        return AlterTableKeyOp(
            table_name=self.table_name,
            key_type=self.existing_key_type,
            key_columns=self.existing_key_columns,
            schema=self.schema,
            existing_key_type=self.key_type,
            existing_key_columns=self.key_columns,
        )

    def __str__(self) -> str:
        return (f"AlterTableKeyOp(table_name={self.table_name!r}, "
               f"key_type={self.key_type!r}, key_columns={self.key_columns!r}, "
               f"schema={self.schema!r}, existing_key_type={self.existing_key_type!r}, "
               f"existing_key_columns={self.existing_key_columns!r})")

@Operations.register_operation("alter_table_partition")
class AlterTablePartitionOp(ops.AlterTableOp):
    """Represent an ALTER TABLE PARTITION BY operation for StarRocks."""

    def __init__(
        self,
        table_name: str,
        partition_method: str,
        schema: Optional[str] = None,
        existing_partition_method: Optional[str] = None,
    ):
        """
        Invoke an ALTER TABLE PARTITION BY operation for StarRocks.
        Args:
            table_name: The name of the table.
            partition_method: The method of the partition, such as 'RANGE(dt)', without pre-created partitions.
            schema: The schema of the table.
        """
        super().__init__(table_name, schema=schema)
        self.partition_method = partition_method
        self.existing_partition_method = existing_partition_method

    @property
    def partition_by(self) -> str:
        """
        Get the partition by string for the ALTER TABLE PARTITION BY operation.
        It DOES NOT include the pre-created partitions.
        Because pre-created partitions should be created with ALTER TABLE ADD PARTITION operation.
        """
        return self.partition_method

    @classmethod
    def alter_table_partition(
        cls,
        operations: Operations,
        table_name: str,
        partition_method: str,
        schema: Optional[str] = None,
        existing_partition_method: Optional[str] = None,
    ):
        """
        Invoke an ALTER TABLE PARTITION BY operation for StarRocks.
        The same as __init__ method.
        """
        op = cls(table_name, partition_method, schema=schema, existing_partition_method=existing_partition_method)
        return operations.invoke(op)

    def reverse(self) -> "AlterTablePartitionOp":
        if self.existing_partition_method is None:
            raise NotImplementedError("Cannot reverse AlterTablePartitionOp without existing_partition_method")
        return AlterTablePartitionOp(
            table_name=self.table_name,
            partition_method=self.existing_partition_method,
            schema=self.schema,
            existing_partition_method=self.partition_method,
        )

    def __str__(self) -> str:
        return (f"AlterTablePartitionOp(table_name={self.table_name!r}, "
               f"partition_method={self.partition_method!r}, schema={self.schema!r}, "
               f"existing_partition_method={self.existing_partition_method!r})")

@Operations.register_operation("alter_table_distribution")
class AlterTableDistributionOp(ops.AlterTableOp):
    """Represent an ALTER TABLE DISTRIBUTED BY operation for StarRocks."""

    def __init__(
        self,
        table_name: str,
        distribution_method: str,
        buckets: Optional[int] = None,
        schema: Optional[str] = None,
        existing_distribution_method: Optional[str] = None,
        existing_buckets: Optional[int] = None,
    ):
        """Invoke an ALTER TABLE DISTRIBUTED BY operation for StarRocks.
        Args:
            table_name: The name of the table.
            distribution_method: The method of the distribution, without BUCKETS.
            buckets: The buckets of the distribution.
            schema: The schema of the table.
        """
        super().__init__(table_name, schema=schema)
        self.distribution_method = distribution_method
        self.buckets = buckets
        self.existing_distribution_method = existing_distribution_method
        self.existing_buckets = existing_buckets

    @property
    def distributed_by(self) -> str:
        """Get the integrated distributed by string for the ALTER TABLE DISTRIBUTED BY operation.
        It includes the BUCKETS if it's not None.
        """
        return f"{self.distribution_method}{f' BUCKETS {self.buckets}' if self.buckets else ''}"

    @classmethod
    def alter_table_distribution(
        cls,
        operations: Operations,
        table_name: str,
        distribution_method: str,
        buckets: Optional[int] = None,
        schema: Optional[str] = None,
        existing_distribution_method: Optional[str] = None,
        existing_buckets: Optional[int] = None,
    ):
        """Invoke an ALTER TABLE DISTRIBUTED BY operation for StarRocks.
        The same as __init__ method.
        """
        op = cls(
            table_name,
            distribution_method,
            buckets,
            schema=schema,
            existing_distribution_method=existing_distribution_method,
            existing_buckets=existing_buckets,
        )
        return operations.invoke(op)

    def reverse(self) -> "AlterTableDistributionOp":
        if self.existing_distribution_method is None:
            raise NotImplementedError("Cannot reverse AlterTableDistributionOp without existing_distribution_method")
        return AlterTableDistributionOp(
            table_name=self.table_name,
            distribution_method=self.existing_distribution_method,
            buckets=self.existing_buckets,
            schema=self.schema,
            existing_distribution_method=self.distribution_method,
            existing_buckets=self.buckets,
        )

    def __str__(self) -> str:
        return (f"AlterTableDistributionOp(table_name={self.table_name!r}, "
               f"distribution_method={self.distribution_method!r}, buckets={self.buckets!r}, "
               f"schema={self.schema!r}, existing_distribution_method={self.existing_distribution_method!r}, "
               f"existing_buckets={self.existing_buckets!r})")


@Operations.register_operation("alter_table_order")
class AlterTableOrderOp(ops.AlterTableOp):
    """Represent an ALTER TABLE ORDER BY operation for StarRocks."""

    def __init__(
        self,
        table_name: str,
        order_by: Union[str, List[str]],
        schema: Optional[str] = None,
        existing_order_by: Optional[str] = None,
    ):
        super().__init__(table_name, schema=schema)
        self.order_by = order_by
        self.existing_order_by = existing_order_by

    @classmethod
    def alter_table_order(
        cls,
        operations: Operations,
        table_name: str,
        order_by: str,
        schema: Optional[str] = None,
        existing_order_by: Optional[str] = None,
    ):
        """Invoke an ALTER TABLE ORDER BY operation for StarRocks."""
        op = cls(table_name, order_by, schema=schema, existing_order_by=existing_order_by)
        return operations.invoke(op)

    def reverse(self) -> "AlterTableOrderOp":
        if self.existing_order_by is None:
            raise NotImplementedError("Cannot reverse AlterTableOrderOp without existing_order_by")
        return AlterTableOrderOp(
            table_name=self.table_name,
            order_by=self.existing_order_by,
            schema=self.schema,
            existing_order_by=self.order_by,
        )

    def __str__(self) -> str:
        return (f"AlterTableOrderOp(table_name={self.table_name!r}, "
               f"order_by={self.order_by!r}, schema={self.schema!r}, "
               f"existing_order_by={self.existing_order_by!r})")


@Operations.register_operation("alter_table_properties")
class AlterTablePropertiesOp(ops.AlterTableOp):
    """Represent an ALTER TABLE SET (...) operation for StarRocks properties."""

    def __init__(
        self,
        table_name: str,
        properties: Dict[str, str],
        schema: Optional[str] = None,
        existing_properties: Optional[Dict[str, str]] = None,
    ):
        super().__init__(table_name, schema=schema)
        self.properties = properties
        self.existing_properties = existing_properties

    @classmethod
    def alter_table_properties(
        cls,
        operations: Operations,
        table_name: str,
        properties: dict,
        schema: Optional[str] = None,
        existing_properties: Optional[Dict[str, str]] = None,
    ):
        """Invoke an ALTER TABLE SET (...) operation for StarRocks properties."""
        op = cls(table_name, properties, schema=schema, existing_properties=existing_properties)
        return operations.invoke(op)

    def reverse(self) -> "AlterTablePropertiesOp":
        if self.existing_properties is None:
            raise NotImplementedError("Cannot reverse AlterTablePropertiesOp without existing_properties")
        return AlterTablePropertiesOp(
            table_name=self.table_name,
            properties=self.existing_properties,
            schema=self.schema,
            existing_properties=self.properties,
        )

    def __str__(self) -> str:
        return (f"AlterTablePropertiesOp(table_name={self.table_name!r}, "
               f"properties={self.properties!r}, schema={self.schema!r}, "
               f"existing_properties={self.existing_properties!r})")



# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation functions for alter_table_xxx, ordered:
# view → mv, alter → create → drop
import logging

from alembic.operations import Operations
from sqlalchemy import MetaData

from starrocks.alembic.ops import (
    AlterMaterializedViewOp,
    AlterTableDistributionOp,
    AlterTableEngineOp,
    AlterTableKeyOp,
    AlterTableOrderOp,
    AlterTablePartitionOp,
    AlterTablePropertiesOp,
    AlterViewOp,
    CreateMaterializedViewOp,
    CreateViewOp,
    DropMaterializedViewOp,
    DropViewOp,
)
from starrocks.sql.ddl import (
    AlterMaterializedView,
    AlterView,
    CreateMaterializedView,
    CreateView,
    DropMaterializedView,
    DropView,
)
from starrocks.sql.schema import MaterializedView, View


logger = logging.getLogger(__name__)


@Operations.implementation_for(AlterViewOp)
def alter_view(operations: Operations, op: AlterViewOp):
    """Execute an ALTER VIEW statement."""
    logger.debug("implementation alter_view: %s", op.view_name)
    view = op.to_view()
    operations.execute(AlterView(view))


@Operations.implementation_for(CreateViewOp)
def create_view(operations: Operations, op: CreateViewOp):
    """Execute a CREATE VIEW statement."""
    logger.debug("implementation create_view: %s", op.view_name)
    view = op.to_view()
    operations.execute(CreateView(view, or_replace=op.or_replace, if_not_exists=op.if_not_exists))


@Operations.implementation_for(DropViewOp)
def drop_view(operations: Operations, op: DropViewOp) -> None:
    """Implementation for the 'drop_view' operation."""
    logger.debug("implementation drop_view: %s", op.view_name)
    # For DROP VIEW, we only need name and schema
    view = op.to_view()
    operations.execute(DropView(view, if_exists=op.if_exists))


@Operations.implementation_for(AlterMaterializedViewOp)
def alter_materialized_view(operations: Operations, op: AlterMaterializedViewOp) -> None:
    """
    Execute ALTER MATERIALIZED VIEW statement.

    Only supports altering mutable attributes: refresh, properties.
    """
    logger.debug("implementation alter_materialized_view: %s", op.view_name)
    operations.execute(AlterMaterializedView(
        mv_name=op.view_name,
        schema=op.schema,
        refresh=op.refresh,
        properties=op.properties,
    ))


@Operations.implementation_for(CreateMaterializedViewOp)
def create_materialized_view(operations: Operations, op: CreateMaterializedViewOp) -> None:
    """Execute a CREATE MATERIALIZED VIEW statement."""
    logger.debug("implementation create_materialized_view: %s", op.view_name)
    mv = op.to_mv()
    operations.execute(CreateMaterializedView(mv, or_replace=op.or_replace, if_not_exists=op.if_not_exists))


@Operations.implementation_for(DropMaterializedViewOp)
def drop_materialized_view(operations: Operations, op: DropMaterializedViewOp) -> None:
    """Implementation for the 'drop_materialized_view' operation."""
    logger.debug("implementation drop_materialized_view: %s", op.view_name)
    # For DROP MATERIALIZED VIEW, we only need name and schema, definition is not required
    # Create a minimal MaterializedView object for DROP operation
    mv = MaterializedView(
        op.view_name,
        MetaData(),
        definition='<placeholder_definition>',  # Empty definition for DROP
        schema=op.schema,
    )
    operations.execute(DropMaterializedView(mv, if_exists=op.if_exists))


# Implementation functions for alter_table_xxx, ordered according to StarRocks grammar:
# engine → key → (comment) → partition → distribution → order by → properties
@Operations.implementation_for(AlterTableEngineOp)
def alter_table_engine(operations, op: AlterTableEngineOp):
    logger.error(
        "ALTER TABLE ENGINE is not currently supported for StarRocks. "
        "Table: %s, Engine: %s", op.table_name, op.engine
    )
    raise NotImplementedError("ALTER TABLE ENGINE is not yet supported")


@Operations.implementation_for(AlterTableKeyOp)
def alter_table_key(operations, op: AlterTableKeyOp):
    logger.error(
        "ALTER TABLE KEY is not currently supported for StarRocks. "
        "Table: %s, Key Type: %s, Columns: %s",
        op.table_name, op.key_type, op.key_columns
    )
    raise NotImplementedError("ALTER TABLE KEY is not yet supported")


@Operations.implementation_for(AlterTablePartitionOp)
def alter_table_partition(operations, op: AlterTablePartitionOp):
    logger.error(
        "ALTER TABLE PARTITION is not currently supported for StarRocks. "
        "Table: %s, Partition: %s", op.table_name, op.partition_method
    )
    raise NotImplementedError("ALTER TABLE PARTITION is not yet supported")


@Operations.implementation_for(AlterTableDistributionOp)
def alter_table_distribution(operations, op: AlterTableDistributionOp):
    from starrocks.sql.ddl import AlterTableDistribution
    operations.execute(
        AlterTableDistribution(
            op.table_name,
            op.distribution_method,
            buckets=op.buckets,
            schema=op.schema
        )
    )


@Operations.implementation_for(AlterTableOrderOp)
def alter_table_order(operations, op: AlterTableOrderOp):
    from starrocks.sql.ddl import AlterTableOrder
    operations.execute(
        AlterTableOrder(op.table_name, op.order_by, schema=op.schema)
    )


@Operations.implementation_for(AlterTablePropertiesOp)
def alter_table_properties(operations, op: AlterTablePropertiesOp):
    from starrocks.sql.ddl import AlterTableProperties
    operations.execute(
        AlterTableProperties(op.table_name, op.properties, schema=op.schema)
    )



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from inspect import isclass
import re
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import sqlalchemy.dialects.mysql.types as mysql_types
from sqlalchemy.engine import Dialect
from sqlalchemy.sql import sqltypes
from sqlalchemy.types import UserDefinedType


class BOOLEAN(sqltypes.BOOLEAN):
    __visit_name__ = "BOOLEAN"


class TINYINT(mysql_types.TINYINT):
    __visit_name__ = "TINYINT"


class SMALLINT(mysql_types.SMALLINT):
    __visit_name__ = "SMALLINT"


class INTEGER(mysql_types.INTEGER):
    __visit_name__ = "INTEGER"


class BIGINT(mysql_types.BIGINT):
    __visit_name__ = "BIGINT"


class LARGEINT(sqltypes.Integer):
    __visit_name__ = "LARGEINT"


class DECIMAL(mysql_types.DECIMAL):
    __visit_name__ = "DECIMAL"


class DOUBLE(mysql_types.DOUBLE):
    __visit_name__ = "DOUBLE"


class FLOAT(mysql_types.FLOAT):
    __visit_name__ = "FLOAT"


class CHAR(mysql_types.CHAR):
    __visit_name__ = "CHAR"


class VARCHAR(mysql_types.VARCHAR):
    __visit_name__ = "VARCHAR"


class STRING(mysql_types.TEXT):
    __visit_name__ = "STRING"


class BINARY(sqltypes.BINARY):
    __visit_name__ = "BINARY"


class VARBINARY(sqltypes.VARBINARY):
    __visit_name__ = "VARBINARY"


class DATETIME(mysql_types.DATETIME):
    __visit_name__ = "DATETIME"

    _reg = re.compile(r"(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)\.?(\d+)?")

    def result_processor(self, dialect: Dialect, coltype: object):
        def process(value):
            if isinstance(value, str):
                m = self._reg.match(value)
                if not m:
                    raise ValueError(
                        "could not parse %r as a datetime value" % (value,)
                    )
                return datetime(*[int(x or 0) for x in m.groups()])
            else:
                return value

        return process

class DATE(sqltypes.DATE):
    __visit_name__ = "DATE"

    def literal_processor(self, dialect: Dialect) -> Callable[[date], str]:
        def process(value: date) -> str:
            return f"TO_DATE('{value}')"

        return process

    _reg = re.compile(r"(\d+)-(\d+)-(\d+)")

    def result_processor(self, dialect: Dialect, coltype: object):
        def process(value):
            if isinstance(value, str):
                m = self._reg.match(value)
                if not m:
                    raise ValueError("could not parse %r as a date value" % (value,))
                return date(*[int(x or 0) for x in m.groups()])
            else:
                return value

        return process


class HLL(sqltypes.Numeric):
    __visit_name__ = "HLL"


class BITMAP(sqltypes.Numeric):
    __visit_name__ = "BITMAP"


class PERCENTILE(sqltypes.Numeric):
    __visit_name__ = "PERCENTILE"


class StructuredType(UserDefinedType):
    @staticmethod
    def _check_subtype(subtype_obj: Union[sqltypes.TypeEngine, type]) -> sqltypes.TypeEngine:
        """
        Check if the subtype is a valid structured type.
        return: an instance of a type, rather than a type itself
        """
        if isclass(subtype_obj):
            if issubclass(subtype_obj, StructuredType):
                raise TypeError(f"'{subtype_obj.__name__}' should be an instance of StructuredType, not a class")
            return subtype_obj()
        return subtype_obj

    def get_col_spec(self, **kw) -> str:
        return "InvalidStructuredType<>"

    def get_sub_type_col_spec(self, sub_type, **kw) -> str:
        if hasattr(sub_type, 'get_col_spec'):
            return sub_type.get_col_spec(**kw)
        else:
            return str(sub_type)

    def get_sub_item_types(self) -> Set[sqltypes.TypeEngine]:
        """
        Get all the sub item types of this structured type recursively.
        Which is need for sqlacodegen to import correct types.
        """
        raise NotImplementedError("get_sub_item_types is not implemented for this pure Structuredtype")


class ARRAY(StructuredType):
    """
    Usage:
        ARRAY(item_type)

    Examples:
        ARRAY(INTEGER)
        ARRAY(ARRAY(STRING))
        ARRAY(STRUCT(name=STRING, address=MAP(STRING, ARRAY(STRING))))
    """

    __visit_name__ = "ARRAY"

    def __init__(self, item_type: sqltypes.TypeEngine, **kwargs):
        self.item_type = self._check_subtype(item_type)
        super().__init__(**kwargs)

    @property
    def python_type(self) -> Optional[Type[List[Any]]]:
        return list

    def __repr__(self):
        return f"ARRAY({repr(self.item_type)})"

    def get_col_spec(self, **kw) -> str:
        inner_type_sql = self.get_sub_type_col_spec(self.item_type)
        return f"ARRAY<{inner_type_sql}>"

    def get_sub_item_types(self) -> Set[sqltypes.TypeEngine]:
        types = {self.item_type}
        if hasattr(self.item_type, 'get_sub_item_types'):
            types.update(self.item_type.get_sub_item_types())
        return types

class MAP(StructuredType):
    """
    Usage:
        MAP(key_type, value_type)

    Examples:
        MAP(INTEGER, STRING)
        MAP(STRING, MAP(INTEGER, STRING))
        MAP(STRING, STRUCT(name=STRING, age=ARRAY(INTEGER)))
    """

    __visit_name__ = "MAP"

    def __init__(self, key_type: sqltypes.TypeEngine, value_type: sqltypes.TypeEngine, **kwargs: Any):
        self.key_type = self._check_subtype(key_type)
        self.value_type = self._check_subtype(value_type)
        super().__init__()

    @property
    def python_type(self) -> Optional[Type[Dict[Any, Any]]]:
        return dict

    def __repr__(self):
        return f"MAP({repr(self.key_type)}, {repr(self.value_type)})"

    def get_col_spec(self, **kw) -> str:
        key_type_sql = self.get_sub_type_col_spec(self.key_type, **kw)
        value_type_sql = self.get_sub_type_col_spec(self.value_type, **kw)
        return f"MAP<{key_type_sql}, {value_type_sql}>"

    def get_sub_item_types(self) -> Set[sqltypes.TypeEngine]:
        types = set[sqltypes.TypeEngine]({self.key_type, self.value_type})
        if hasattr(self.key_type, 'get_sub_item_types'):
            types.update(self.key_type.get_sub_item_types())
        if hasattr(self.value_type, 'get_sub_item_types'):
            types.update(self.value_type.get_sub_item_types())
        return types

class STRUCT(StructuredType):
    """
    Usage:
        STRUCT((name, type), (name, type), ..., name=type, ...)

    Examples:
        STRUCT((name, STRING), (age, INTEGER))
        STRUCT(name=STRING, info=STRUCT(age=INTEGER, city=STRING))
        STRUCT(name=STRING, address=MAP(STRING, ARRAY(STRING)))
"""

    __visit_name__ = "STRUCT"

    def __init__(self, *fields: Tuple[STRING, sqltypes.TypeEngine], **kwfields: sqltypes.TypeEngine):
        self.field_tuples = tuple(
            (name, self._check_subtype(type_))
            for (name, type_) in (fields + tuple(kwfields.items()))
        )
        self._field_dict = {
            name.lower(): type_ for (name, type_) in self.field_tuples
        }
        super().__init__()

    @property
    def python_type(self) -> Optional[Type[Any]]:
        return None

    def __repr__(self):
        fields = ", ".join(
            f"{name}={repr(type_)}" for name, type_ in self.field_tuples
        )
        return f"STRUCT({fields})"

    def get_col_spec(self, **kw) -> str:
        fields_sql = []
        for name, type_ in self.field_tuples:
            type_sql = self.get_sub_type_col_spec(type_, **kw)
            fields_sql.append(f"{name} {type_sql}")
        return f"STRUCT<{', '.join(fields_sql)}>"

    def get_sub_item_types(self) -> Set[sqltypes.TypeEngine]:
        types = set[sqltypes.TypeEngine]()
        for _, type_ in self.field_tuples:
            types.add(type_)
            if hasattr(type_, 'get_sub_item_types'):
                types.update(type_.get_sub_item_types())
        return types


class JSON(sqltypes.JSON):
    __visit_name__ = "JSON"
